# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2007-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


from pyx import reader

class char_info_word:

    def __init__(self, word):
        self.width_index  = int((word & 0xFF000000) >> 24) #make sign-safe
        self.height_index = (word & 0x00F00000) >> 20
        self.depth_index  = (word & 0x000F0000) >> 16
        self.italic_index = (word & 0x0000FC00) >> 10
        self.tag          = (word & 0x00000300) >> 8
        self.remainder    = (word & 0x000000FF)


class TFMfile:

    def __init__(self, file, debug=0):
        with reader.bytesreader(file.read()) as file:

            #
            # read pre header
            #

            self.lf = file.readint16()
            self.lh = file.readint16()
            self.bc = file.readint16()
            self.ec = file.readint16()
            self.nw = file.readint16()
            self.nh = file.readint16()
            self.nd = file.readint16()
            self.ni = file.readint16()
            self.nl = file.readint16()
            self.nk = file.readint16()
            self.ne = file.readint16()
            self.np = file.readint16()

            if not (self.bc-1 <= self.ec <= 255 and
                    self.ne <= 256 and
                    self.lf == 6+self.lh+(self.ec-self.bc+1)+self.nw+self.nh+self.nd
                    +self.ni+self.nl+self.nk+self.ne+self.np):
                raise RuntimeError("error in TFM pre-header")

            if debug:
                print("lh=%d" % self.lh)

            #
            # read header
            #

            self.checksum = file.readint32()
            self.designsize = file.readint32()
            assert self.designsize > 0, "invald design size"
            if self.lh > 2:
                assert self.lh > 11, "inconsistency in TFM file: incomplete field"
                self.charcoding = file.readstring(40)
            else:
                self.charcoding = None

            if self.lh > 12:
                assert self.lh > 16, "inconsistency in TFM file: incomplete field"
                self.fontfamily = file.readstring(20)
            else:
                self.fontfamily = None

            if debug:
                print("(FAMILY %s)" % self.fontfamily)
                print("(CODINGSCHEME %s)" % self.charcoding)
                print("(DESINGSIZE R %f)" % (16.0*self.designsize/16777216))

            if self.lh > 17:
                self.sevenbitsave = file.readuchar()
                # ignore the following two bytes
                file.readint16()
                facechar = file.readuchar()
                # decode ugly face specification into the Knuth suggested string
                if facechar < 18:
                    if facechar >= 12:
                        self.face = "E"
                        facechar -= 12
                    elif facechar >= 6:
                        self.face = "C"
                        facechar -= 6
                    else:
                        self.face = "R"

                    if facechar >= 4:
                        self.face = "L" + self.face
                        facechar -= 4
                    elif facechar >= 2:
                        self.face = "B" + self.face
                        facechar -= 2
                    else:
                        self.face = "M" + self.face

                    if facechar == 1:
                        self.face = self.face[0] + "I" + self.face[1]
                    else:
                        self.face = self.face[0] + "R" + self.face[1]

                else:
                    self.face = None
            else:
                self.sevenbitsave = self.face = None

            if self.lh > 18:
                # just ignore the rest
                print(file.read((self.lh-18)*4))

            #
            # read char_info
            #

            self.char_info = [None]*(self.ec+1)
            for charcode in range(self.bc, self.ec+1):
                self.char_info[charcode] = char_info_word(file.readint32())
                if self.char_info[charcode].width_index == 0:
                    # disable character if width_index is zero
                    self.char_info[charcode] = None

            #
            # read widths
            #

            self.width = [None for width_index in range(self.nw)]
            for width_index in range(self.nw):
                self.width[width_index] = file.readint32()

            #
            # read heights
            #

            self.height = [None for height_index in range(self.nh)]
            for height_index in range(self.nh):
                self.height[height_index] = file.readint32()

            #
            # read depths
            #

            self.depth = [None for depth_index in range(self.nd)]
            for depth_index in range(self.nd):
                self.depth[depth_index] = file.readint32()

            #
            # read italic
            #

            self.italic = [None for italic_index in range(self.ni)]
            for italic_index in range(self.ni):
                self.italic[italic_index] = file.readint32()

            #
            # read lig_kern
            #

            # XXX decode to lig_kern_command

            self.lig_kern = [None for lig_kern_index in range(self.nl)]
            for lig_kern_index in range(self.nl):
                self.lig_kern[lig_kern_index] = file.readint32()

            #
            # read kern
            #

            self.kern = [None for kern_index in range(self.nk)]
            for kern_index in range(self.nk):
                self.kern[kern_index] = file.readint32()

            #
            # read exten
            #

            # XXX decode to extensible_recipe

            self.exten = [None for exten_index in range(self.ne)]
            for exten_index in range(self.ne):
                self.exten[exten_index] = file.readint32()

            #
            # read param
            #

            # XXX decode

            self.param = [None for param_index in range(self.np)]
            for param_index in range(self.np):
                self.param[param_index] = file.readint32()



from pyx import *

g = graph.graphxy(width=8)
g.plot(graph.data.file("errorbar.dat", x=1, y=2, dy=3),
       [graph.style.symbol(), graph.style.errorbar()])
g.writeEPSfile("errorbar")
g.writePDFfile("errorbar")
g.writeSVGfile("errorbar")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2003-2004,2006,2007 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import io, logging, math, re, string, struct, sys
from pyx import  bbox, canvas, color, epsfile, config, path, reader, trafo, unit
from . import texfont, tfmfile

logger = logging.getLogger("pyx")


_DVI_CHARMIN     =   0 # typeset a character and move right (range min)
_DVI_CHARMAX     = 127 # typeset a character and move right (range max)
_DVI_SET1234     = 128 # typeset a character and move right
_DVI_SETRULE     = 132 # typeset a rule and move right
_DVI_PUT1234     = 133 # typeset a character
_DVI_PUTRULE     = 137 # typeset a rule
_DVI_NOP         = 138 # no operation
_DVI_BOP         = 139 # beginning of page
_DVI_EOP         = 140 # ending of page
_DVI_PUSH        = 141 # save the current positions (h, v, w, x, y, z)
_DVI_POP         = 142 # restore positions (h, v, w, x, y, z)
_DVI_RIGHT1234   = 143 # move right
_DVI_W0          = 147 # move right by w
_DVI_W1234       = 148 # move right and set w
_DVI_X0          = 152 # move right by x
_DVI_X1234       = 153 # move right and set x
_DVI_DOWN1234    = 157 # move down
_DVI_Y0          = 161 # move down by y
_DVI_Y1234       = 162 # move down and set y
_DVI_Z0          = 166 # move down by z
_DVI_Z1234       = 167 # move down and set z
_DVI_FNTNUMMIN   = 171 # set current font (range min)
_DVI_FNTNUMMAX   = 234 # set current font (range max)
_DVI_FNT1234     = 235 # set current font
_DVI_SPECIAL1234 = 239 # special (dvi extention)
_DVI_FNTDEF1234  = 243 # define the meaning of a font number
_DVI_PRE         = 247 # preamble
_DVI_POST        = 248 # postamble beginning
_DVI_POSTPOST    = 249 # postamble ending

_DVI_VERSION     = 2   # dvi version

# position variable indices
_POS_H           = 0
_POS_V           = 1
_POS_W           = 2
_POS_X           = 3
_POS_Y           = 4
_POS_Z           = 5

# reader states
_READ_PRE       = 1
_READ_NOPAGE    = 2
_READ_PAGE      = 3
_READ_POST      = 4 # XXX not used
_READ_POSTPOST  = 5 # XXX not used
_READ_DONE      = 6


class DVIError(Exception): pass


class DVIfile:

    def __init__(self, filename, debug=0, debugfile=sys.stdout):
        """ opens the dvi file and reads the preamble """
        self.filename = filename
        self.debug = debug
        self.debugfile = debugfile
        self.debugstack = []

        self.fonts = {}
        self.activefont = None

        # stack of fonts and fontscale currently used (used for VFs)
        self.fontstack = []
        self.stack = []

        # pointer to currently active page
        self.actpage = None

        # stack for self.file, self.fonts and self.stack, needed for VF inclusion
        self.statestack = []

        self.file = reader.reader(self.filename)

        # currently read byte in file (for debugging output)
        self.filepos = None

        self._read_pre()

    # helper routines

    def beginsubpage(self, attrs):
        c = canvas.canvas(attrs)
        c.parent = self.actpage
        c.markers = {}
        self.actpage.insert(c)
        self.actpage = c

    def endsubpage(self):
        for key, value in list(self.actpage.markers.items()):
            self.actpage.parent.markers[key] = self.actpage.trafo.apply(*value)
        self.actpage = self.actpage.parent

    def flushtext(self, fontmap):
        """ finish currently active text object """
        if self.activetext:
            x, y, charcodes = self.activetext
            x_pt, y_pt = x * self.pyxconv, -y*self.pyxconv
            self.actpage.insert(self.activefont.text_pt(x_pt, y_pt, charcodes, fontmap=fontmap))
            if self.debug:
                self.debugfile.write("[%s]\n" % "".join([chr(char) for char in self.activetext[2]]))
            self.activetext = None

    def putrule(self, height, width, advancepos, fontmap):
        self.flushtext(fontmap)
        x1 =  self.pos[_POS_H] * self.pyxconv
        y1 = -self.pos[_POS_V] * self.pyxconv
        w = width * self.pyxconv
        h = height * self.pyxconv

        if height > 0 and width > 0:
            if self.debug:
                self.debugfile.write("%d: %srule height %d, width %d (???x??? pixels)\n" %
                                     (self.filepos, advancepos and "set" or "put", height, width))
            self.actpage.fill(path.rect_pt(x1, y1, w, h))
        else:
            if self.debug:
                self.debugfile.write("%d: %srule height %d, width %d (invisible)\n" %
                                     (self.filepos, advancepos and "set" or "put", height, width))

        if advancepos:
            if self.debug:
                self.debugfile.write(" h:=%d+%d=%d, hh:=???\n" %
                                     (self.pos[_POS_H], width, self.pos[_POS_H]+width))
            self.pos[_POS_H] += width * self.scale

    def putchar(self, char, advancepos, id1234, fontmap):
        dx = advancepos and self.activefont.getwidth_dvi(char) or 0

        if self.debug:
            self.debugfile.write("%d: %s%s%d h:=%d+%d=%d, hh:=???\n" %
                                 (self.filepos,
                                  advancepos and "set" or "put",
                                  id1234 and "%i " % id1234 or "char",
                                  char,
                                  self.pos[_POS_H], dx, self.pos[_POS_H]+dx))

        if isinstance(self.activefont, texfont.virtualfont):
            # virtual font handling
            afterpos = list(self.pos)
            afterpos[_POS_H] += dx
            self._push_dvistring(self.activefont.getchar(char), self.activefont.getfonts(), afterpos,
                                 self.activefont.getsize_pt(), fontmap)
        else:
            if self.activetext is None:
                self.activetext = (self.pos[_POS_H], self.pos[_POS_V], [])
            self.activetext[2].append(char)
            self.pos[_POS_H] += dx

        if (not advancepos) or self.singlecharmode:
            self.flushtext(fontmap)

    def usefont(self, fontnum, id1234, fontmap):
        self.flushtext(fontmap)
        self.activefont = self.fonts[fontnum]
        if self.debug:
            self.debugfile.write("%d: fnt%s%i current font is %s\n" %
                                 (self.filepos,
                                  id1234 and "%i " % id1234 or "num",
                                  fontnum,
                                  self.fonts[fontnum].name))


    def definefont(self, cmdnr, num, c, q, d, fontname):
        # cmdnr: type of fontdef command (only used for debugging output)
        # c:     checksum
        # q:     scaling factor (fix_word)
        #        Note that q is actually s in large parts of the documentation.
        # d:     design size (fix_word)

        # check whether it's a virtual font by trying to open it. if this fails, it is an ordinary TeX font
        try:
            with config.open(fontname, [config.format.vf]) as fontfile:
                afont = texfont.virtualfont(fontname, fontfile, c, q/self.tfmconv, d/self.tfmconv, self.tfmconv, self.pyxconv, self.debug>1)
        except EnvironmentError:
            afont = texfont.TeXfont(fontname, c, q/self.tfmconv, d/self.tfmconv, self.tfmconv, self.pyxconv, self.debug>1)

        self.fonts[num] = afont

        if self.debug:
            self.debugfile.write("%d: fntdef%d %i: %s\n" % (self.filepos, cmdnr, num, fontname))

#            scale = round((1000.0*self.conv*q)/(self.trueconv*d))
#            m = 1.0*q/d
#            scalestring = scale!=1000 and " scaled %d" % scale or ""
#            print ("Font %i: %s%s---loaded at size %d DVI units" %
#                   (num, fontname, scalestring, q))
#            if scale!=1000:
#                print " (this font is magnified %d%%)" % round(scale/10)

    def special(self, s, fontmap):
        x =  self.pos[_POS_H] * self.pyxconv
        y = -self.pos[_POS_V] * self.pyxconv
        if self.debug:
            self.debugfile.write("%d: xxx '%s'\n" % (self.filepos, s))
        if s == "header=l3backend-dvips.pro":
            # ignore header included in every file by l3backend since about texlive 2020
            return
        if not s.startswith("PyX:"):
            logger.warning("ignoring special '%s'" % s)
            return

        # it is in general not safe to continue using the currently active font because
        # the specials may involve some gsave/grestore operations
        self.flushtext(fontmap)

        command, args = s[4:].split()[0], s[4:].split()[1:]
        if command == "color_begin":
            if args[0] == "cmyk":
                c = color.cmyk(float(args[1]), float(args[2]), float(args[3]), float(args[4]))
            elif args[0] == "gray":
                c = color.gray(float(args[1]))
            elif args[0] == "hsb":
                c = color.hsb(float(args[1]), float(args[2]), float(args[3]))
            elif args[0] == "rgb":
                c = color.rgb(float(args[1]), float(args[2]), float(args[3]))
            elif args[0] == "RGB":
                c = color.rgb(int(args[1])/255.0, int(args[2])/255.0, int(args[3])/255.0)
            elif args[0] == "texnamed":
                try:
                    c = getattr(color.cmyk, args[1])
                except AttributeError:
                    raise RuntimeError("unknown TeX color '%s', aborting" % args[1])
            elif args[0] == "pyxcolor":
                # pyx.color.cmyk.PineGreen or
                # pyx.color.cmyk(0,0,0,0.0)
                pat = re.compile(r"(pyx\.)?(color\.)?(?P<model>(cmyk)|(rgb)|(grey)|(gray)|(hsb))[\.]?(?P<arg>.*)")
                sd = pat.match(" ".join(args[1:]))
                if sd:
                    sd = sd.groupdict()
                    if sd["arg"][0] == "(":
                        numpat = re.compile(r"[+-]?((\d+\.\d*)|(\d*\.\d+)|(\d+))([eE][+-]\d+)?")
                        arg = tuple([float(x[0]) for x in numpat.findall(sd["arg"])])
                        try:
                            c = getattr(color, sd["model"])(*arg)
                        except TypeError or AttributeError:
                            raise RuntimeError("cannot access PyX color '%s' in TeX, aborting" % " ".join(args[1:]))
                    else:
                        try:
                            c = getattr(getattr(color, sd["model"]), sd["arg"])
                        except AttributeError:
                            raise RuntimeError("cannot access PyX color '%s' in TeX, aborting" % " ".join(args[1:]))
                else:
                    raise RuntimeError("cannot access PyX color '%s' in TeX, aborting" % " ".join(args[1:]))
            else:
                raise RuntimeError("color model '%s' cannot be handled by PyX, aborting" % args[0])

            self.beginsubpage([c])
        elif command == "color_end":
            self.endsubpage()
        elif command == "rotate_begin":
            self.beginsubpage([trafo.rotate_pt(float(args[0]), x, y)])
        elif command == "rotate_end":
            self.endsubpage()
        elif command == "scale_begin":
            self.beginsubpage([trafo.scale_pt(float(args[0]), float(args[1]), x, y)])
        elif command == "scale_end":
            self.endsubpage()
        elif command == "epsinclude":
            # parse arguments
            argdict = {}
            for arg in args:
                name, value = arg.split("=")
                argdict[name] = value

            # construct kwargs for epsfile constructor
            epskwargs = {}
            epskwargs["filename"] = argdict["file"]
            epskwargs["bbox"] = bbox.bbox_pt(float(argdict["llx"]), float(argdict["lly"]),
                                           float(argdict["urx"]), float(argdict["ury"]))
            if "width" in argdict:
                epskwargs["width"] = float(argdict["width"]) * unit.t_pt
            if "height" in argdict:
                epskwargs["height"] = float(argdict["height"]) * unit.t_pt
            if "clip" in argdict:
               epskwargs["clip"] = int(argdict["clip"])
            self.actpage.insert(epsfile.epsfile(x * unit.t_pt, y * unit.t_pt, **epskwargs))
        elif command == "marker":
            if len(args) != 1:
                raise RuntimeError("marker contains spaces")
            for c in args[0]:
                if c not in string.ascii_letters + string.digits + "@":
                    raise RuntimeError("marker contains invalid characters")
            if args[0] in self.actpage.markers:
                raise RuntimeError("marker name occurred several times")
            self.actpage.markers[args[0]] = x * unit.t_pt, y * unit.t_pt
        else:
            raise RuntimeError("unknown PyX special '%s', aborting" % command)

    # routines for pushing and popping different dvi chunks on the reader

    def _push_dvistring(self, dvi, fonts, afterpos, fontsize, fontmap):
        """ push dvi string with defined fonts on top of reader
        stack. Every positions gets scaled relatively by the factor
        scale. After interpretating the dvi chunk, continue with self.pos=afterpos.
        The designsize of the virtual font is passed as a fix_word

        """

        #if self.debug:
        #    self.debugfile.write("executing new dvi chunk\n")
        self.debugstack.append(self.debug)
        self.debug = 0

        self.statestack.append((self.file, self.fonts, self.activefont, afterpos, self.stack, self.scale))

        # units in vf files are relative to the size of the font and given as fix_words
        # which can be converted to floats by diving by 2**20.
        # This yields the following scale factor for the height and width of rects:
        self.scale = fontsize/2**20/self.pyxconv

        self.file = reader.bytesreader(dvi)
        self.fonts = fonts
        self.stack = []
        self.filepos = 0

        self.usefont(0, 0, fontmap)

    def _pop_dvistring(self, fontmap):
        self.flushtext(fontmap)
        #if self.debug:
        #    self.debugfile.write("finished executing dvi chunk\n")
        self.debug = self.debugstack.pop()

        self.file.close()
        self.file, self.fonts, self.activefont, self.pos, self.stack, self.scale = self.statestack.pop()

    # routines corresponding to the different reader states of the dvi maschine

    def _read_pre(self):
        afile = self.file
        while True:
            self.filepos = afile.tell()
            cmd = afile.readuchar()
            if cmd == _DVI_NOP:
                pass
            elif cmd == _DVI_PRE:
                if afile.readuchar() != _DVI_VERSION: raise DVIError
                num = afile.readuint32()
                den = afile.readuint32()
                self.mag = afile.readuint32()

                # For the interpretation of the lengths in dvi and tfm files, 
                # three conversion factors are relevant:
                # - self.tfmconv: tfm units -> dvi units
                # - self.pyxconv: dvi units -> (PostScript) points
                # - self.conv:    dvi units -> pixels
                self.tfmconv = (25400000.0/num)*(den/473628672.0)/16.0

                # calculate conv as described in the DVIType docu using 
                # a given resolution in dpi
                self.resolution = 300.0
                self.conv = (num/254000.0)*(self.resolution/den)

                # self.pyxconv is the conversion factor from the dvi units
                # to (PostScript) points. It consists of
                # - self.mag/1000.0:   magstep scaling
                # - self.conv:         conversion from dvi units to pixels
                # - 1/self.resolution: conversion from pixels to inch
                # - 72               : conversion from inch to points
                self.pyxconv = self.mag/1000.0*self.conv/self.resolution*72

                # scaling used for rules when VF chunks are interpreted
                self.scale = 1

                comment = afile.read(afile.readuchar())
                return
            else:
                raise DVIError

    def readpage(self, pageid=None, fontmap=None, singlecharmode=False, attrs=[]):
        """ reads a page from the dvi file

        This routine reads a page from the dvi file which is
        returned as a canvas. When there is no page left in the
        dvifile, None is returned and the file is closed properly."""

        self.singlecharmode = singlecharmode

        while True:
            self.filepos = self.file.tell()
            cmd = self.file.readuchar()
            if cmd == _DVI_NOP:
                pass
            elif cmd == _DVI_BOP:
                ispageid = [self.file.readuint32() for i in range(10)]
                if pageid is not None and ispageid != pageid:
                    raise DVIError("invalid pageid")
                if self.debug:
                    self.debugfile.write("%d: beginning of page %i\n" % (self.filepos, ispageid[0]))
                self.file.readuint32()
                break
            elif cmd == _DVI_POST:
                self.file.close()
                return None # nothing left
            else:
                raise DVIError

        self.actpage = canvas.canvas(attrs)
        self.actpage.markers = {}
        self.pos = [0, 0, 0, 0, 0, 0]

        # tuple (hpos, vpos, codepoints) to be output, or None if no output is pending
        self.activetext = None

        while True:
            afile = self.file
            self.filepos = afile.tell()
            try:
                cmd = afile.readuchar()
            except struct.error:
                # we most probably (if the dvi file is not corrupt) hit the end of a dvi chunk,
                # so we have to continue with the rest of the dvi file
                self._pop_dvistring(fontmap)
                continue
            if cmd == _DVI_NOP:
                pass
            if cmd >= _DVI_CHARMIN and cmd <= _DVI_CHARMAX:
                self.putchar(cmd, True, 0, fontmap)
            elif cmd >= _DVI_SET1234 and cmd < _DVI_SET1234 + 4:
                self.putchar(afile.readint(cmd - _DVI_SET1234 + 1), True, cmd-_DVI_SET1234+1, fontmap)
            elif cmd == _DVI_SETRULE:
                self.putrule(afile.readint32()*self.scale, afile.readint32()*self.scale, True, fontmap)
            elif cmd >= _DVI_PUT1234 and cmd < _DVI_PUT1234 + 4:
                self.putchar(afile.readint(cmd - _DVI_PUT1234 + 1), False, cmd-_DVI_SET1234+1, fontmap)
            elif cmd == _DVI_PUTRULE:
                self.putrule(afile.readint32()*self.scale, afile.readint32()*self.scale, False, fontmap)
            elif cmd == _DVI_EOP:
                self.flushtext(fontmap)
                if self.debug:
                    self.debugfile.write("%d: eop\n \n" % self.filepos)
                return self.actpage
            elif cmd == _DVI_PUSH:
                self.stack.append(list(self.pos))
                if self.debug:
                    self.debugfile.write("%s: push\n"
                                         "level %d:(h=%d,v=%d,w=%d,x=%d,y=%d,z=%d,hh=???,vv=???)\n" %
                                         ((self.filepos, len(self.stack)-1) + tuple(self.pos)))
            elif cmd == _DVI_POP:
                self.flushtext(fontmap)
                self.pos = self.stack.pop()
                if self.debug:
                    self.debugfile.write("%s: pop\n"
                                         "level %d:(h=%d,v=%d,w=%d,x=%d,y=%d,z=%d,hh=???,vv=???)\n" %
                                         ((self.filepos, len(self.stack)) + tuple(self.pos)))
            elif cmd >= _DVI_RIGHT1234 and cmd < _DVI_RIGHT1234 + 4:
                self.flushtext(fontmap)
                dh = afile.readint(cmd - _DVI_RIGHT1234 + 1, 1) * self.scale
                if self.debug:
                    self.debugfile.write("%d: right%d %d h:=%d%+d=%d, hh:=???\n" %
                                         (self.filepos,
                                          cmd - _DVI_RIGHT1234 + 1,
                                          dh,
                                          self.pos[_POS_H],
                                          dh,
                                          self.pos[_POS_H]+dh))
                self.pos[_POS_H] += dh
            elif cmd == _DVI_W0:
                self.flushtext(fontmap)
                if self.debug:
                    self.debugfile.write("%d: w0 %d h:=%d%+d=%d, hh:=???\n" %
                                         (self.filepos,
                                          self.pos[_POS_W],
                                          self.pos[_POS_H],
                                          self.pos[_POS_W],
                                          self.pos[_POS_H]+self.pos[_POS_W]))
                self.pos[_POS_H] += self.pos[_POS_W]
            elif cmd >= _DVI_W1234 and cmd < _DVI_W1234 + 4:
                self.flushtext(fontmap)
                self.pos[_POS_W] = afile.readint(cmd - _DVI_W1234 + 1, 1) * self.scale
                if self.debug:
                    self.debugfile.write("%d: w%d %d h:=%d%+d=%d, hh:=???\n" %
                                         (self.filepos,
                                          cmd - _DVI_W1234 + 1,
                                          self.pos[_POS_W],
                                          self.pos[_POS_H],
                                          self.pos[_POS_W],
                                          self.pos[_POS_H]+self.pos[_POS_W]))
                self.pos[_POS_H] += self.pos[_POS_W]
            elif cmd == _DVI_X0:
                self.flushtext(fontmap)
                if self.debug:
                    self.debugfile.write("%d: x0 %d h:=%d%+d=%d, hh:=???\n" %
                                         (self.filepos,
                                          self.pos[_POS_X],
                                          self.pos[_POS_H],
                                          self.pos[_POS_X],
                                          self.pos[_POS_H]+self.pos[_POS_X]))
                self.pos[_POS_H] += self.pos[_POS_X]
            elif cmd >= _DVI_X1234 and cmd < _DVI_X1234 + 4:
                self.flushtext(fontmap)
                self.pos[_POS_X] = afile.readint(cmd - _DVI_X1234 + 1, 1) * self.scale
                if self.debug:
                    self.debugfile.write("%d: x%d %d h:=%d%+d=%d, hh:=???\n" %
                                         (self.filepos,
                                          cmd - _DVI_X1234 + 1,
                                          self.pos[_POS_X],
                                          self.pos[_POS_H],
                                          self.pos[_POS_X],
                                          self.pos[_POS_H]+self.pos[_POS_X]))
                self.pos[_POS_H] += self.pos[_POS_X]
            elif cmd >= _DVI_DOWN1234 and cmd < _DVI_DOWN1234 + 4:
                self.flushtext(fontmap)
                dv = afile.readint(cmd - _DVI_DOWN1234 + 1, 1) * self.scale
                if self.debug:
                    self.debugfile.write("%d: down%d %d v:=%d%+d=%d, vv:=???\n" %
                                         (self.filepos,
                                          cmd - _DVI_DOWN1234 + 1,
                                          dv,
                                          self.pos[_POS_V],
                                          dv,
                                          self.pos[_POS_V]+dv))
                self.pos[_POS_V] += dv
            elif cmd == _DVI_Y0:
                self.flushtext(fontmap)
                if self.debug:
                    self.debugfile.write("%d: y0 %d v:=%d%+d=%d, vv:=???\n" %
                                         (self.filepos,
                                          self.pos[_POS_Y],
                                          self.pos[_POS_V],
                                          self.pos[_POS_Y],
                                          self.pos[_POS_V]+self.pos[_POS_Y]))
                self.pos[_POS_V] += self.pos[_POS_Y]
            elif cmd >= _DVI_Y1234 and cmd < _DVI_Y1234 + 4:
                self.flushtext(fontmap)
                self.pos[_POS_Y] = afile.readint(cmd - _DVI_Y1234 + 1, 1) * self.scale
                if self.debug:
                    self.debugfile.write("%d: y%d %d v:=%d%+d=%d, vv:=???\n" %
                                         (self.filepos,
                                          cmd - _DVI_Y1234 + 1,
                                          self.pos[_POS_Y],
                                          self.pos[_POS_V],
                                          self.pos[_POS_Y],
                                          self.pos[_POS_V]+self.pos[_POS_Y]))
                self.pos[_POS_V] += self.pos[_POS_Y]
            elif cmd == _DVI_Z0:
                self.flushtext(fontmap)
                if self.debug:
                    self.debugfile.write("%d: z0 %d v:=%d%+d=%d, vv:=???\n" %
                                         (self.filepos,
                                          self.pos[_POS_Z],
                                          self.pos[_POS_V],
                                          self.pos[_POS_Z],
                                          self.pos[_POS_V]+self.pos[_POS_Z]))
                self.pos[_POS_V] += self.pos[_POS_Z]
            elif cmd >= _DVI_Z1234 and cmd < _DVI_Z1234 + 4:
                self.flushtext(fontmap)
                self.pos[_POS_Z] = afile.readint(cmd - _DVI_Z1234 + 1, 1) * self.scale
                if self.debug:
                    self.debugfile.write("%d: z%d %d v:=%d%+d=%d, vv:=???\n" %
                                         (self.filepos,
                                          cmd - _DVI_Z1234 + 1,
                                          self.pos[_POS_Z],
                                          self.pos[_POS_V],
                                          self.pos[_POS_Z],
                                          self.pos[_POS_V]+self.pos[_POS_Z]))
                self.pos[_POS_V] += self.pos[_POS_Z]
            elif cmd >= _DVI_FNTNUMMIN and cmd <= _DVI_FNTNUMMAX:
                self.usefont(cmd - _DVI_FNTNUMMIN, 0, fontmap)
            elif cmd >= _DVI_FNT1234 and cmd < _DVI_FNT1234 + 4:
                # note that according to the DVI docs, for four byte font numbers,
                # the font number is signed. Don't ask why!
                fntnum = afile.readint(cmd - _DVI_FNT1234 + 1, cmd == _DVI_FNT1234 + 3)
                self.usefont(fntnum, cmd-_DVI_FNT1234+1, fontmap)
            elif cmd >= _DVI_SPECIAL1234 and cmd < _DVI_SPECIAL1234 + 4:
                self.special(afile.read(afile.readint(cmd - _DVI_SPECIAL1234 + 1)).decode("ascii"), fontmap)
            elif cmd >= _DVI_FNTDEF1234 and cmd < _DVI_FNTDEF1234 + 4:
                if cmd == _DVI_FNTDEF1234:
                    num = afile.readuchar()
                elif cmd == _DVI_FNTDEF1234+1:
                    num = afile.readuint16()
                elif cmd == _DVI_FNTDEF1234+2:
                    num = afile.readuint24()
                elif cmd == _DVI_FNTDEF1234+3:
                    # Cool, here we have according to docu a signed int. Why?
                    num = afile.readint32()
                self.definefont(cmd-_DVI_FNTDEF1234+1,
                                num,
                                afile.readint32(),
                                afile.readint32(),
                                afile.readint32(),
                                afile.read(afile.readuchar()+afile.readuchar()).decode("ascii"))
            else:
                raise DVIError



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2012 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2002-2012 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

"""The canvas module provides a PostScript canvas class and related classes

A canvas holds a collection of all elements and corresponding attributes to be
displayed. """

import io, logging, os, sys, string, tempfile
from . import attr, baseclasses, config, document, style, trafo, svgwriter, unit
from . import bbox as bboxmodule

logger = logging.getLogger("pyx")

def _wrappedindocument(method):
    def wrappedindocument(self, file=None, **kwargs):
        page_kwargs = {}
        write_kwargs = {}
        for name, value in list(kwargs.items()):
            if name.startswith("page_"):
                page_kwargs[name[5:]] = value
            elif name.startswith("write_"):
                write_kwargs[name[6:]] = value
            else:
                logger.warning("implicit page keyword argument passing is deprecated; keyword argument '%s' of %s method should be changed to 'page_%s'" %
                               (name, method.__name__, name))
                page_kwargs[name] = value
        d = document.document([document.page(self, **page_kwargs)])
        self.__name__ = method.__name__
        self.__doc__ = method.__doc__
        return method(d, file, **write_kwargs)
    return wrappedindocument

#
# clipping class
#

class SVGclippath(svgwriter.SVGresource):

    def __init__(self, path):
        self.svgid = "clippath%d" % id(path)
        super().__init__("clip-path", self.svgid)
        self.path = path

    def output(self, xml, writer, registry):
        xml.startSVGElement("clipPath", {"id": self.svgid})
        # TODO: clip-rule missing (defaults to nonzero)
        xml.startSVGElement("path", {"d": self.path.returnSVGdata()})
        xml.endSVGElement("path")
        xml.endSVGElement("clipPath")


class clip(attr.attr):

    """class for use in canvas constructor which clips to a path"""

    def __init__(self, path):
        """construct a clip instance for a given path"""
        self.path = path

    def processPS(self, file, writer, context, registry):
        file.write("newpath\n")
        self.path.outputPS(file, writer)
        file.write("clip\n")

    def processPDF(self, file, writer, context, registry):
        self.path.outputPDF(file, writer)
        file.write("W n\n")

    def processSVGattrs(self, attrs, writer, context, registry):
        clippath = SVGclippath(self.path)
        registry.add(clippath)
        attrs["clip-path"] = "url(#%s)" % clippath.svgid


#
# general canvas class
#

class canvas(baseclasses.canvasitem):

    """a canvas holds a collection of canvasitems"""

    def __init__(self, attrs=None, textengine=None, ipython_bboxenlarge=1*unit.t_pt):

        """construct a canvas

        The canvas can be modfied by supplying a list of attrs, which have
        to be instances of one of the following classes:
         - trafo.trafo (leading to a global transformation of the canvas)
         - canvas.clip (clips the canvas)
         - style.strokestyle, style.fillstyle (sets some global attributes of the canvas)

        Note that, while the first two properties are fixed for the
        whole canvas, the last one can be changed via canvas.set().

        The textengine instance used for the text method can be specified
        using the textengine argument. It defaults to text.defaulttextengine

        """

        self.items = []
        self.trafo = trafo.identity
        self.clip = None
        self.layers = {}
        if attrs is None:
            attrs = []
        if textengine is not None:
            self.textengine = textengine
        else:
            # prevent cyclic imports
            from . import text
            self.textengine = text.defaulttextengine
        self.ipython_bboxenlarge = ipython_bboxenlarge

        attr.checkattrs(attrs, [trafo.trafo_pt, clip, style.style])
        attrs = attr.mergeattrs(attrs)
        self.modifies_state = bool(attrs)

        self.styles = attr.getattrs(attrs, [style.style])

        # trafos (and one possible clip operation) are applied from left to
        # right in the attrs list -> reverse for calculating total trafo
        for aattr in reversed(attr.getattrs(attrs, [trafo.trafo_pt, clip])):
            if isinstance(aattr, trafo.trafo_pt):
                self.trafo = self.trafo * aattr
            else:
                if self.clip is not None:
                    raise ValueError("single clipping allowed only")
                self.clip = clip(aattr.path.transformed(self.trafo))

    def __len__(self):
        return len(self.items)

    def __getitem__(self, i):
        return self.items[i]

    def clear(self):
        """clear canvas"""
        self.items.clear()
        self.layers.clear()

    def _repr_png_(self):
        """
        Automatically represent as PNG graphic when evaluated in IPython notebook.
        """
        return self.pipeGS(device="png16m", page_bboxenlarge=self.ipython_bboxenlarge).getvalue()

    def _repr_svg_(self):
        """
        Automatically represent as SVG graphic when evaluated in IPython notebook.
        """
        f = io.BytesIO()
        self.writeSVGfile(f, page_bboxenlarge=self.ipython_bboxenlarge)
        return f.getvalue().decode("utf-8")

    def bbox(self):
        """returns bounding box of canvas

        Note that this bounding box doesn't take into account the linewidths, so
        is less accurate than the one used when writing the output to a file.
        """
        obbox = bboxmodule.empty()
        for cmd in self.items:
            obbox += cmd.bbox()

        # transform according to our global transformation and
        # intersect with clipping bounding box (which has already been
        # transformed in canvas.__init__())
        obbox.transform(self.trafo)
        if self.clip is not None:
            obbox *= self.clip.path.bbox()
        return obbox

    def processPS(self, file, writer, context, registry, bbox):
        context = context()
        if self.items:
            if self.modifies_state:
                file.write("gsave\n")
                for attr in self.styles:
                    attr.processPS(file, writer, context, registry)
                if self.clip is not None:
                    self.clip.processPS(file, writer, context, registry)
                if self.trafo is not trafo.identity:
                    self.trafo.processPS(file, writer, context, registry)
            nbbox = bboxmodule.empty()
            for item in self.items:
                item.processPS(file, writer, context, registry, nbbox)
            # update bounding bbox
            nbbox.transform(self.trafo)
            if self.clip is not None:
                nbbox *= self.clip.path.bbox()
            bbox += nbbox
            if self.modifies_state:
                file.write("grestore\n")

    def processPDF(self, file, writer, context, registry, bbox):
        context = context()
        textregion = False
        context.trafo = context.trafo * self.trafo
        if self.items:
            if self.modifies_state:
                file.write("q\n") # gsave
                for attr in self.styles:
                    if isinstance(attr, style.fillstyle):
                        context.fillstyles.append(attr)
                    attr.processPDF(file, writer, context, registry)
                if self.clip is not None:
                    self.clip.processPDF(file, writer, context, registry)
                if self.trafo is not trafo.identity:
                    self.trafo.processPDF(file, writer, context, registry)
            nbbox = bboxmodule.empty()
            for item in self.items:
                if not writer.textaspath:
                    if item.requiretextregion():
                        if not textregion:
                            file.write("BT\n")
                            textregion = True
                    else:
                        if textregion:
                            file.write("ET\n")
                            textregion = False
                            context.selectedfont = None
                item.processPDF(file, writer, context, registry, nbbox)
            if textregion:
                file.write("ET\n")
                textregion = False
                context.selectedfont = None
            # update bounding bbox
            nbbox.transform(self.trafo)
            if self.clip is not None:
                nbbox *= self.clip.path.bbox()
            bbox += nbbox
            if self.modifies_state:
                file.write("Q\n") # grestore

    def processSVG(self, xml, writer, context, registry, bbox):
        if self.items:
            if self.modifies_state:
                context = context()
                attrs = {}
                for attr in self.styles:
                    attr.processSVGattrs(attrs, writer, context, registry)
                if self.clip is not None:
                    self.clip.processSVGattrs(attrs, writer, context, registry)
                    if self.trafo is not trafo.identity:
                        # trafo needs to be applied after clipping
                        # thus write g and start anew
                        xml.startSVGElement("g", attrs)
                        attrs = {}
                        self.trafo.processSVGattrs(attrs, writer, context, registry)
                elif self.trafo is not trafo.identity:
                    self.trafo.processSVGattrs(attrs, writer, context, registry)
                xml.startSVGElement("g", attrs)
            nbbox = bboxmodule.empty()
            for item in self.items:
                item.processSVG(xml, writer, context, registry, nbbox)
            # update bounding bbox
            nbbox.transform(self.trafo)
            if self.clip is not None:
                nbbox *= self.clip.path.bbox()
            bbox += nbbox
            if self.modifies_state:
                xml.endSVGElement("g")
                if self.clip is not None and self.trafo is not trafo.identity:
                    xml.endSVGElement("g")

    def layer(self, name, above=None, below=None):
        """create or get a layer with name

        A layer is a canvas itself and can be used to combine drawing
        operations for ordering purposes, i.e., what is above and below each
        other. The layer name is a dotted string, where dots are used to form
        a hierarchy of layer groups. When inserting a layer, it is put on top
        of its layer group except when another layer of this group is specified
        by means of the parameters above or below.

        """
        if above is not None and below is not None:
            raise ValueError("above and below cannot be specified at the same time")
        try:
            group, layer = name.split(".", 1)
        except ValueError:
            if name in self.layers:
                if above is not None or below is not None:
                    # remove for repositioning
                    self.items.remove(self.layers[name])
            else:
                # create new layer
                self.layers[name] = canvas(textengine=self.textengine)
                if above is None and below is None:
                    self.items.append(self.layers[name])

            # (re)position layer
            if above is not None:
                self.items.insert(self.items.index(self.layers[above])+1, self.layers[name])
            elif below is not None:
                self.items.insert(self.items.index(self.layers[below]), self.layers[name])

            return self.layers[name]
        else:
            if not group in self.layers:
                self.layers[group] = self.insert(canvas(textengine=self.textengine))
            if above is not None:
                abovegroup, above = above.split(".", 1)
                assert abovegroup == group
            if below is not None:
                belowgroup, below = below.split(".", 1)
                assert belowgroup == group
            return self.layers[group].layer(layer, above=above, below=below)

    def insert(self, item, attrs=None):
        """insert item in the canvas.

        If attrs are passed, a canvas containing the item is inserted applying
        attrs. If replace is not None, the new item is
        positioned accordingly in the canvas.

        returns the item, possibly wrapped in a canvas

        """

        if not isinstance(item, baseclasses.canvasitem):
            raise ValueError("only instances of baseclasses.canvasitem can be inserted into a canvas")

        if attrs:
            sc = canvas(attrs)
            sc.insert(item)
            item = sc

        self.items.append(item)
        return item

    def draw(self, path, attrs):
        """draw path on canvas using the style given by args

        The argument attrs consists of PathStyles, which modify
        the appearance of the path, PathDecos, which add some new
        visual elements to the path, or trafos, which are applied
        before drawing the path.

        """
        from . import deco
        attrs = attr.mergeattrs(attrs)
        attr.checkattrs(attrs, [deco.deco, baseclasses.deformer, style.style])

        for adeformer in attr.getattrs(attrs, [baseclasses.deformer]):
            path = adeformer.deform(path)

        styles = attr.getattrs(attrs, [style.style])
        dp = deco.decoratedpath(path, styles=styles)

        # add path decorations and modify path accordingly
        for adeco in attr.getattrs(attrs, [deco.deco]):
            adeco.decorate(dp, self.textengine)

        self.insert(dp)

    def stroke(self, path, attrs=[]):
        """stroke path on canvas using the style given by args

        The argument attrs consists of PathStyles, which modify
        the appearance of the path, PathDecos, which add some new
        visual elements to the path, or trafos, which are applied
        before drawing the path.

        """
        from . import deco
        self.draw(path, [deco.stroked]+list(attrs))

    def fill(self, path, attrs=[]):
        """fill path on canvas using the style given by args

        The argument attrs consists of PathStyles, which modify
        the appearance of the path, PathDecos, which add some new
        visual elements to the path, or trafos, which are applied
        before drawing the path.

        """
        from . import deco
        self.draw(path, [deco.filled]+list(attrs))

    def settextengine(self, textengine):
        """sets the textengine to be used to within the text and text_pt methods"""

        self.textengine = textengine

    def text(self, x, y, atext, *args, **kwargs):
        """insert a text into the canvas

        inserts a textbox created by self.textengine.text into the canvas

        returns the inserted textbox"""

        return self.insert(self.textengine.text(x, y, atext, *args, **kwargs))


    def text_pt(self, x, y, atext, *args):
        """insert a text into the canvas

        inserts a textbox created by self.textengine.text_pt into the canvas

        returns the inserted textbox"""

        return self.insert(self.textengine.text_pt(x, y, atext, *args))

    writeEPSfile = _wrappedindocument(document.document.writeEPSfile)
    writePSfile = _wrappedindocument(document.document.writePSfile)
    writePDFfile = _wrappedindocument(document.document.writePDFfile)
    writeSVGfile = _wrappedindocument(document.document.writeSVGfile)
    writetofile = _wrappedindocument(document.document.writetofile)


    def _gscmd(self, device, filename, resolution=100, gs="gs", gsoptions=[],
               textalphabits=4, graphicsalphabits=4, ciecolor=False, **kwargs):

        cmd = [gs, "-dEPSCrop", "-dNOPAUSE", "-dQUIET", "-dBATCH", "-r%d" % resolution, "-sDEVICE=%s" % device, "-sOutputFile=%s" % filename]
        if textalphabits is not None:
            cmd.append("-dTextAlphaBits=%i" % textalphabits)
        if graphicsalphabits is not None:
            cmd.append("-dGraphicsAlphaBits=%i" % graphicsalphabits)
        if ciecolor:
            cmd.append("-dUseCIEColor")
        cmd.extend(gsoptions)

        return cmd, kwargs

    def writeGSfile(self, filename=None, device=None, input="eps", **kwargs):
        """
        convert EPS or PDF output to a file via Ghostscript

        If filename is None it is auto-guessed from the script name. If
        filename is "-", the output is written to stdout. In both cases, a
        device needs to be specified to define the format.

        If device is None, but a filename with suffix is given, PNG files will
        be written using the png16m device and JPG files using the jpeg device.
        """
        if filename is None:
            if not sys.argv[0].endswith(".py"):
                raise RuntimeError("could not auto-guess filename")
            if device.startswith("png"):
                filename = sys.argv[0][:-2] + "png"
            elif device.startswith("jpeg"):
                filename = sys.argv[0][:-2] + "jpg"
            else:
                filename = sys.argv[0][:-2] + device
        if device is None:
            if filename.endswith(".png"):
                device = "png16m"
            elif filename.endswith(".jpg"):
                device = "jpeg"
            else:
                raise RuntimeError("could not auto-guess device")

        cmd, kwargs = self._gscmd(device, filename, **kwargs)

        if input == "eps":
            cmd.append("-")
            p = config.Popen(cmd, stdin=config.PIPE)
            self.writeEPSfile(p.stdin, **kwargs)
            p.stdin.close()
            p.wait()
        elif input == "pdf":
            # PDF files need to be accesible by random access and thus we need to create
            # a temporary file
            with tempfile.NamedTemporaryFile("wb", delete=False) as f:
                self.writePDFfile(f, **kwargs)
                fname = f.name
            cmd.append(fname)
            config.Popen(cmd).wait()
            os.unlink(fname)
        else:
            raise RuntimeError("input 'eps' or 'pdf' expected")


    def pipeGS(self, device, input="eps", **kwargs):
        """
        returns a BytesIO instance with the Ghostscript output of the EPS or PDF
        """

        cmd, kwargs = self._gscmd(device, "-", **kwargs)

        with tempfile.NamedTemporaryFile("wb", delete=False) as f:
            if input == "eps":
                self.writeEPSfile(f, **kwargs)
            elif input == "pdf":
                self.writePDFfile(f, **kwargs)
            else:
                raise RuntimeError("input 'eps' or 'pdf' expected")
            fname = f.name

        cmd.append(fname)
        p = config.Popen(cmd, stdout=config.PIPE)
        data, error = p.communicate()
        os.unlink(fname)

        if error:
            raise ValueError("error received while waiting for ghostscript")
        return io.BytesIO(data)



from pyx import *

def mark(x, y):
    return path.circle(x, y, 0.1)

c = canvas.canvas()

p1 = path.curve(0, 0, 1, 0, 1, 1, 2, 1)
c.stroke(p1)
c.fill(mark(*p1.atbegin()))
c.fill(mark(*p1.at(0.5*p1.arclen())))
c.fill(mark(*p1.atend()))

p2 = path.curve(3, 0, 4, 0, 4, 1, 5, 1)
c.stroke(p2)
c.fill(mark(*p2.at(p2.begin()+0.5)))
c.fill(mark(*p2.at(p2.end()-0.5)))

c.writeEPSfile("at")
c.writePDFfile("at")
c.writeSVGfile("at")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2004, 2022 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2003-2004 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2012, 2022 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


from pyx import unit, box
from pyx.graph.axis import tick


# rater
# conceptual remarks:
# - raters are used to calculate a rating for a realization of something
# - a rating means a positive floating point value
# - ratings are used to order those realizations by their suitability
#   (small ratings are better)
# - a rating of None means not suitable at all (those realizations should be
#   thrown out)


class cube:
    """a value rater
    - a cube rater has an optimal value, where the rating becomes zero
    - for a left (below the optimum) and a right value (above the optimum),
      the rating is value is set to 1 (modified by an overall weight factor
      for the rating)
    - the analytic form of the rating is cubic for both, the left and
      the right side of the rater, independently"""

    def __init__(self, opt, left=None, right=None, weight=1):
        """initializes the rater
        - by default, left is set to zero, right is set to 3*opt
        - left should be smaller than opt, right should be bigger than opt
        - weight should be positive and is a factor multiplicated to the rates"""
        if left is None:
            left = 0
        if right is None:
            right = 3*opt
        self.opt = opt
        self.left = left
        self.right = right
        self.weight = weight

    def rate(self, value, density):
        """returns a rating for a value
        - the density lineary rescales the rater (the optimum etc.),
          e.g. a value bigger than one increases the optimum (when it is
          positive) and a value lower than one decreases the optimum (when
          it is positive); the density itself should be positive"""
        opt = self.opt * density
        if value < opt:
            other = self.left * density
        elif value > opt:
            other = self.right * density
        else:
            return 0
        factor = (value - opt) / float(other - opt)
        return self.weight * (factor ** 3)


class distance:
    # TODO: update docstring
    """a distance rater (rates a list of distances)
    - the distance rater rates a list of distances by rating each independently
      and returning the average rate
    - there is an optimal value, where the rate becomes zero
    - the analytic form is linary for values above the optimal value
      (twice the optimal value has the rating one, three times the optimal
      value has the rating two, etc.)
    - the analytic form is reciprocal subtracting one for values below the
      optimal value (halve the optimal value has the rating one, one third of
      the optimal value has the rating two, etc.)"""

    def __init__(self, opt, weight=0.1):
        """inititializes the rater
        - opt is the optimal length (a visual PyX length)
        - weight should be positive and is a factor multiplicated to the rates"""
        self.opt = opt
        self.weight = weight

    def rate(self, distances, density):
        """rate distances
        - the distances are a list of positive floats in PostScript points
        - the density lineary rescales the rater (the optimum etc.),
          e.g. a value bigger than one increases the optimum (when it is
          positive) and a value lower than one decreases the optimum (when
          it is positive); the density itself should be positive"""
        if len(distances):
            opt = unit.topt(self.opt) / density
            rate = 0
            for distance in distances:
                if distance < opt:
                    rate += self.weight * (opt / distance - 1)
                else:
                    rate += self.weight * (distance / opt - 1)
            return rate / float(len(distances))


class rater:
    """a rater for ticks
    - the rating of axes is splited into two separate parts:
      - rating of the ticks in terms of the number of ticks, subticks,
        labels, etc.
      - rating of the label distances
    - in the end, a rate for ticks is the sum of these rates
    - it is useful to first just rate the number of ticks etc.
      and selecting those partitions, where this fits well -> as soon
      as an complete rate (the sum of both parts from the list above)
      of a first ticks is below a rate of just the number of ticks,
      subticks labels etc. of other ticks, those other ticks will never
      be better than the first one -> we gain speed by minimizing the
      number of ticks, where label distances have to be taken into account)
    - both parts of the rating are shifted into instances of raters
      defined above --- right now, there is not yet a strict interface
      for this delegation (should be done as soon as it is needed)"""

    def __init__(self, ticks, labels, range, distance):
        """initializes the axis rater
        - ticks and labels are lists of instances of a value rater
        - the first entry in ticks rate the number of ticks, the
          second the number of subticks, etc.; when there are no
          ticks of a level or there is not rater for a level, the
          level is just ignored
        - labels is analogous, but for labels
        - within the rating, all ticks with a higher level are
          considered as ticks for a given level
        - range is a value rater instance, which rates the covering
          of an axis range by the ticks (as a relative value of the
          tick range vs. the axis range), ticks might cover less or
          more than the axis range (for the standard automatic axis
          partition schemes an extention of the axis range is normal
          and should get some penalty)
        - distance is an distance rater instance"""
        self.ticks = ticks
        self.labels = labels
        self.range = range
        self.distance = distance

    def __call__(self, **kwargs):
        return rater(**utils.merge_members_kwargs(self, [kwargs],
                                                 ["ticks", "labels", "range", "distance"]))

    def rateticks(self, axis, ticks, density):
        """rates ticks by the number of ticks, subticks, labels etc.
        - takes into account the number of ticks, subticks, labels
          etc. and the coverage of the axis range by the ticks
        - when there are no ticks of a level or there was not rater
          given in the constructor for a level, the level is just
          ignored
        - the method returns the sum of the rating results divided
          by the sum of the weights of the raters
        - within the rating, all ticks with a higher level are
          considered as ticks for a given level"""
        maxticklevel, maxlabellevel = tick.maxlevels(ticks)
        if not maxticklevel and not maxlabellevel:
            return None
        numticks = [0]*maxticklevel
        numlabels = [0]*maxlabellevel
        for t in ticks:
            if t.ticklevel is not None:
                for level in range(t.ticklevel, maxticklevel):
                    numticks[level] += 1
            if t.labellevel is not None:
                for level in range(t.labellevel, maxlabellevel):
                    numlabels[level] += 1
        rate = 0
        weight = 0
        for numtick, rater in zip(numticks, self.ticks):
            rate += rater.rate(numtick, density)
            weight += rater.weight
        for numlabel, rater in zip(numlabels, self.labels):
            rate += rater.rate(numlabel, density)
            weight += rater.weight
        return rate/weight

    def raterange(self, tickrange, datarange):
        """rate the range covered by the ticks compared to the range
        of the data
        - tickrange and datarange are the ranges covered by the ticks
          and the data in graph coordinates
        - usually, the datarange is 1 (ticks are calculated for a
          given datarange)
        - the ticks might cover less or more than the data range (for
          the standard automatic axis partition schemes an extention
          of the axis range is normal and should get some penalty)"""
        return self.range.rate(tickrange, datarange)

    def ratelayout(self, axiscanvas, density):
        """rate distances of the labels in an axis canvas
        - the distances should be collected as box distances of
          subsequent labels
        - the axiscanvas provides a labels attribute for easy
          access to the labels whose distances have to be taken
          into account
        - the density is used within the distancerate instance"""
        if axiscanvas.labels is None: # to disable any layout rating
            return 0
        if len(axiscanvas.labels) > 1:
            try:
                distances = [axiscanvas.labels[i].boxdistance_pt(axiscanvas.labels[i+1])
                             for i in range(len(axiscanvas.labels) - 1)]
            except box.BoxCrossError:
                return None
            return self.distance.rate(distances, density)
        else:
            return None


class linear(rater):
    """a rater with predefined constructor arguments suitable for a linear axis"""

    def __init__(self, ticks=[cube(4), cube(10, weight=0.5)],
                       labels=[cube(4)],
                       range=cube(1, weight=2),
                       distance=distance(1*unit.v_cm)):
        rater.__init__(self, ticks, labels, range, distance)

lin = linear


class logarithmic(rater):
    """a rater with predefined constructor arguments suitable for a logarithmic axis"""

    def __init__(self, ticks=[cube(5, right=20), cube(20, right=100, weight=0.5)],
                       labels=[cube(5, right=20), cube(5, right=20, weight=0.5)],
                       range=cube(1, weight=2),
                       distance=distance(1*unit.v_cm)):
        rater.__init__(self, ticks, labels, range, distance)

log = logarithmic



from pyx import *

clippath = path.circle(0, 0, 1)
drawpath = path.line(-2, -2, 1.2, 2)

cl = canvas.canvas([canvas.clip(clippath)])
cl.stroke(drawpath, [color.rgb.red, style.linewidth(1.0)])

c = canvas.canvas()
c.stroke(drawpath, [style.linewidth(1.0)])
c.insert(cl)
c.stroke(clippath)

c.writeEPSfile("clipping")
c.writePDFfile("clipping")
c.writeSVGfile("clipping")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2004-2006 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


import datetime
from pyx.graph import style
from pyx.graph.axis import axis, rater

"""some experimental code for creating a time axis
- it needs python 2.3 to be used (it is based on the new datetime data type)
- a timeaxis is always based on the datetime data type (there is no distinction between times and dates)
"""

class timeaxis(axis.linear):
    "time axis mapping based "

    # TODO: how to deal with reversed timeaxis?

    def __init__(self, parter=None, rater=rater.linear(), **args):
        super().__init__(self, divisor=None, **args)
        self.parter = parter
        self.rater = rater

    def convert(self, data, x):
        # XXX float division of timedelta instances
        def mstimedelta(td):
            "return the timedelta in microseconds"
            return td.microseconds + 1000000*(td.seconds + 3600*24*td.days)
        return mstimedelta(x - data.min) / float(mstimedelta(data.max - data.min))
        # we could store float(mstimedelta(self.dx)) instead of self.dx, but
        # I prefer a different solution (not based on huge integers) for the
        # future

    zero = datetime.timedelta(0)


class timetick(datetime.datetime):

    # TODO: http://stackoverflow.com/questions/399022/why-cant-i-subclass-datetime-date
    # possible fix: make the datetime an attribute of the tick

    def __new__(cls, year, month, day, ticklevel=0, labellevel=0, label=None, labelattrs=[], **kwargs):
        return datetime.datetime.__new__(cls, year, month, day, **kwargs)

    def __init__(self, year, month, day, ticklevel=0, labellevel=0, label=None, labelattrs=[], **kwargs):
        self.ticklevel = ticklevel
        self.labellevel = labellevel
        self.label = label
        self.labelattrs = labelattrs[:]

    def merge(self, other):
        if self.ticklevel is None or (other.ticklevel is not None and other.ticklevel < self.ticklevel):
            self.ticklevel = other.ticklevel
        if self.labellevel is None or (other.labellevel is not None and other.labellevel < self.labellevel):
            self.labellevel = other.labellevel


class timetexter:

    def __init__(self, format="%c"):
        self.format = format

    def labels(self, ticks):
        for tick in ticks:
            if tick.labellevel is not None and tick.label is None:
                tick.label = tick.strftime(self.format)






# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2007-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2007-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import io, logging, os.path, re
from pyx import font, config
from pyx.font import t1file, afmfile, pfmfile
from pyx.dvi import encfile

logger = logging.getLogger("pyx")

class UnsupportedFontFormat(Exception):
    pass

class UnsupportedPSFragment(Exception):
    pass

class ParseError(Exception):
    pass

_marker = object()

class MAPline:

    tokenpattern = re.compile(r'"(.*?)("\s+|"$|$)|(.*?)(\s+|$)')

    def __init__(self, s):
        """ construct font mapping from line s of font mapping file """
        self.texname = self.basepsname = self.fontfilename = None

        # standard encoding
        self.encodingfilename = None

        # supported postscript fragments occuring in psfonts.map
        # XXX extendfont not yet implemented
        self.reencodefont = self.extendfont = self.slant = None

        # cache for openend font and encoding
        self._font = None
        self._encoding = _marker

        tokens = []
        while len(s):
            match = self.tokenpattern.match(s)
            if match:
                if match.groups()[0] is not None:
                    tokens.append('"%s"' % match.groups()[0])
                else:
                    tokens.append(match.groups()[2])
                s = s[match.end():]
            else:
                raise ParseError("Cannot tokenize string '%s'" % s)

        next_token_is_encfile = False
        for token in tokens:
            if next_token_is_encfile:
                self.encodingfilename = token
                next_token_is_encfile = False
            elif token.startswith("<"):
                if token == "<":
                    next_token_is_encfile = True
                elif token.startswith("<<"):
                    # XXX: support non-partial download here
                    self.fontfilename = token[2:]
                elif token.startswith("<["):
                    self.encodingfilename = token[2:]
                elif token.endswith(".pfa") or token.endswith(".pfb"):
                    self.fontfilename = token[1:]
                elif token.endswith(".enc"):
                    self.encodingfilename = token[1:]
                elif token.endswith(".ttf"):
                    raise UnsupportedFontFormat("TrueType font")
                elif token.endswith(".t42"):
                    raise UnsupportedFontFormat("Type 42 font")
                else:
                    # Assume (as in pdftex's mapfile.c) that we are dealing with a Type 1 font.
                    # Note that this case has in particular appeared with the MinLibertine fonts,
                    # for which this heuristics seems to work well.
                    self.fontfilename = token[1:]
            elif token.startswith('"'):
                pscode = token[1:-1].split()
                # parse standard postscript code fragments
                while pscode:
                    try:
                        arg, cmd = pscode[:2]
                    except Exception:
                        raise UnsupportedPSFragment("Unsupported Postscript fragment '%s'" % pscode)
                    pscode = pscode[2:]
                    if cmd == "ReEncodeFont":
                        self.reencodefont = arg
                    elif cmd == "ExtendFont":
                        self.extendfont = arg
                    elif cmd == "SlantFont":
                        self.slant = float(arg)
                    else:
                        raise UnsupportedPSFragment("Unsupported Postscript fragment '%s %s'" % (arg, cmd))
            else:
                if self.texname is None:
                    self.texname = token
                else:
                    self.basepsname = token
        if self.basepsname is None:
            self.basepsname = self.texname

    def getfontname(self):
        return self.basepsname

    def getfont(self):
        if self._font is None:
            if self.fontfilename is not None:
                with config.open(self.fontfilename, [config.format.type1]) as fontfile:
                    t1font = t1file.T1File.from_PF_bytes(fontfile.read())
                assert self.basepsname == t1font.name, "corrupt MAP file"
                try:
                    with config.open(os.path.splitext(self.fontfilename)[0], [config.format.afm], ascii=True) as metricfile:
                        self._font = font.T1font(t1font, afmfile.AFMfile(metricfile))
                except EnvironmentError:
                    try:
                        # fallback by using the pfm instead of the afm font metric
                        # (in all major TeX distributions there is no pfm file format defined by kpsewhich, but
                        # we can use the type1 format and search for the file including the expected suffix)
                        with config.open("%s.pfm" % os.path.splitext(self.fontfilename)[0], [config.format.type1]) as metricfile:
                            self._font = font.T1font(t1font, pfmfile.PFMfile(metricfile, t1font))
                    except EnvironmentError:
                        # we need to continue without any metric file
                        self._font = font.T1font(t1font)
            else:
                # builtin font
                with config.open(self.basepsname, [config.format.afm], ascii=True) as metricfile:
                    self._font = font.T1builtinfont(self.basepsname, afmfile.AFMfile(metricfile))
        return self._font

    def getencoding(self):
        if self._encoding is _marker:
            if self.encodingfilename is not None:
                with config.open(self.encodingfilename, [config.format.tex_ps_header]) as encodingfile:
                    ef = encfile.ENCfile(encodingfile.read().decode("ascii", errors="surrogateescape"))
                assert ef.name == "/%s" % self.reencodefont
                self._encoding = ef.vector

            else:
                self._encoding = None
        return self._encoding

    def __str__(self):
        return ("'%s' is '%s' read from '%s' encoded as '%s'" %
                (self.texname, self.basepsname, self.fontfile, repr(self.encodingfile)))

# generate fontmap

def readfontmap(filenames):
    """ read font map from filename (without path) """
    fontmap = {}
    for filename in filenames:
        with config.open(filename, [config.format.fontmap, config.format.dvips_config], ascii=True) as mapfile:
            lineno = 0
            for line in mapfile.readlines():
                lineno += 1
                line = line.rstrip()
                if not (line=="" or line[0] in (" ", "%", "*", ";" , "#")):
                    try:
                        fm = MAPline(line)
                    except (ParseError, UnsupportedPSFragment) as e:
                        logger.warning("Ignoring line %i in mapping file '%s': %s" % (lineno, filename, e))
                    except UnsupportedFontFormat as e:
                        pass
                    else:
                        fontmap[fm.texname] = fm
    return fontmap



from pyx import *

g = graph.graphxy(width=8, x=graph.axis.bar())
g.plot(graph.data.file("minimal.dat", xname=0, y=2), [graph.style.bar()])
g.writeEPSfile("minimal")
g.writePDFfile("minimal")
g.writeSVGfile("minimal")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2002-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import io, logging, math
from . import attr, canvas, path, pdfwriter, pswriter, svgwriter, style, unit, trafo
from . import writer as writermodule
from . import bbox as bboxmodule

logger = logging.getLogger("pyx")

class _marker: pass

# TODO: pattern should not derive from canvas but wrap a canvas

class pattern(canvas.canvas, attr.exclusiveattr, style.fillstyle):

    def __init__(self, painttype=1, tilingtype=1, xstep=None, ystep=None,
                 bbox=None, trafo=None, bboxenlarge=None, **kwargs):
        kwargs['attrs'] = attr.mergeattrs([style.linewidth.normal] + kwargs.get('attrs', []))
        canvas.canvas.__init__(self, **kwargs)
        attr.exclusiveattr.__init__(self, pattern)
        self.id = "pattern%d" % id(self)
        self.patterntype = 1
        if painttype not in (1, 2):
            raise ValueError("painttype must be 1 or 2")
        self.painttype = painttype
        if tilingtype not in (1, 2, 3):
            raise ValueError("tilingtype must be 1, 2, or 3")
        self.tilingtype = tilingtype
        self.xstep = xstep
        self.ystep = ystep
        self.patternbbox = bbox
        self.patterntrafo = trafo
        self.bboxenlarge = bboxenlarge

    def __call__(self, painttype=_marker, tilingtype=_marker, xstep=_marker, ystep=_marker,
                 bbox=_marker, trafo=_marker, bboxenlarge=_marker):
        if painttype is not _marker:
            painttype = self.painttype
        if tilingtype is not _marker:
            tilingtype = self.tilingtype
        if xstep is not _marker:
            xstep = self.xstep
        if ystep is not _marker:
            ystep = self.ystep
        if bbox is not _marker:
            bbox = self.bbox
        if trafo is not _marker:
            trafo = self.trafo
        if bboxenlarge is not _marker:
            bboxenlarge = self.bboxenlarge
        return pattern(painttype, tilingtype, xstep, ystep, bbox, trafo, bboxenlarge)

    def processPS(self, file, writer, context, registry):
        # process pattern, letting it register its resources and calculate the bbox of the pattern
        patternfile = writermodule.writer(io.BytesIO())
        realpatternbbox = bboxmodule.empty()
        canvas.canvas.processPS(self, patternfile, writer, pswriter.context(), registry, realpatternbbox)
        patternproc = patternfile.file.getvalue()

        if self.xstep is None:
            xstep = unit.topt(realpatternbbox.width())
        else:
            xstep = unit.topt(self.xstep)
        if self.ystep is None:
            ystep = unit.topt(realpatternbbox.height())
        else:
            ystep = unit.topt(self.ystep)
        if not xstep:
            raise ValueError("xstep in pattern cannot be zero")
        if not ystep:
            raise ValueError("ystep in pattern cannot be zero")
        if self.patternbbox:
            patternbbox = self.patternbbox
        else:
            patternbbox = realpatternbbox
            if self.bboxenlarge:
                patternbbox.enlarge(self.bboxenlarge)

        patternprefix = "\n".join(("<<",
                                   "/PatternType %d" % self.patterntype,
                                   "/PaintType %d" % self.painttype,
                                   "/TilingType %d" % self.tilingtype,
                                   "/BBox [%g %g %g %g]" % patternbbox.highrestuple_pt(),
                                   "/XStep %g" % xstep,
                                   "/YStep %g" % ystep,
                                   "/PaintProc {\nbegin\n"))
        patterntrafostring = self.patterntrafo is None and "matrix" or str(self.patterntrafo)
        patternsuffix = "end\n} bind\n>>\n%s\nmakepattern" % patterntrafostring

        registry.add(pswriter.PSdefinition(self.id, patternprefix.encode("ascii") + patternproc + patternsuffix.encode("ascii")))

        # activate pattern
        file.write("%s setpattern\n" % self.id)

    def processPDF(self, file, writer, context, registry):
        # we need to keep track of the resources used by the pattern, hence
        # we create our own registry, which we merge immediately in the main registry
        patternregistry = pdfwriter.PDFregistry()

        patternfile = writermodule.writer(io.BytesIO())
        realpatternbbox = bboxmodule.empty()
        canvas.canvas.processPDF(self, patternfile, writer, pdfwriter.context(), patternregistry, realpatternbbox)
        patternproc = patternfile.file.getvalue()

        registry.mergeregistry(patternregistry)

        if self.xstep is None:
           xstep = unit.topt(realpatternbbox.width())
        else:
           xstep = unit.topt(self.xstep)
        if self.ystep is None:
            ystep = unit.topt(realpatternbbox.height())
        else:
           ystep = unit.topt(self.ystep)
        if not xstep:
            raise ValueError("xstep in pattern cannot be zero")
        if not ystep:
            raise ValueError("ystep in pattern cannot be zero")
        if self.patternbbox:
            patternbbox = self.patternbbox
        else:
            patternbbox = realpatternbbox
            if self.bboxenlarge:
                patternbbox.enlarge(self.bboxenlarge)
        patterntrafo = self.patterntrafo or trafo.trafo()

        registry.add(PDFpattern(self.id, self.patterntype, self.painttype, self.tilingtype,
                                patternbbox, xstep, ystep, patterntrafo, patternproc, writer, registry, patternregistry))

        # activate pattern
        if context.colorspace != "Pattern":
            # we only set the fill color space (see next comment)
            file.write("/Pattern cs\n")
            context.colorspace = "Pattern"
        if context.strokeattr:
            # using patterns as stroke colors doesn't seem to work, so
            # we just don't do this...
            logger.warning("ignoring stroke color for patterns in PDF")
        if context.fillattr:
            file.write("/%s scn\n"% self.id)

    def processSVGattrs(self, attrs, writer, context, registry):
        assert self.patterntype == 1
        if self.painttype != 1:
            raise ValueError("grayscale patterns not supported")
        # tilingtype is an implementation detail in PS and PDF and is ignored
        if self.xstep is not None or self.ystep is not None or self.bboxenlarge is not None:
            raise ValueError("step and bbox modifications not supported")

        svgpattern = SVGpattern(self)
        registry.add(svgpattern)
        if context.strokeattr:
            context.strokecolor = "url(#%s)" % svgpattern.svgid
        if context.fillattr:
            context.fillcolor = "url(#%s)" % svgpattern.svgid


pattern.clear = attr.clearclass(pattern)


_base = 0.1 * unit.v_cm

class hatched(pattern):
    def __init__(self, dist, angle, strokestyles=[]):
        pattern.__init__(self, painttype=1, tilingtype=1, xstep=dist, ystep=100*unit.t_pt, bbox=None, trafo=trafo.rotate(angle))
        self.strokestyles = attr.mergeattrs([style.linewidth.THIN] + strokestyles)
        attr.checkattrs(self.strokestyles, [style.strokestyle])
        self.dist = dist
        self.angle = angle
        self.stroke(path.line_pt(0, -50, 0, 50), self.strokestyles)

    def __call__(self, dist=None, angle=None, strokestyles=None):
        if dist is None:
            dist = self.dist
        if angle is None:
            angle = self.angle
        if strokestyles is None:
            strokestyles = self.strokestyles
        return hatched(dist, angle, strokestyles)

hatched0 = hatched(_base, 0)
hatched0.SMALL = hatched0(_base/math.sqrt(64))
hatched0.SMALL = hatched0(_base/math.sqrt(64))
hatched0.SMALl = hatched0(_base/math.sqrt(32))
hatched0.SMAll = hatched0(_base/math.sqrt(16))
hatched0.SMall = hatched0(_base/math.sqrt(8))
hatched0.Small = hatched0(_base/math.sqrt(4))
hatched0.small = hatched0(_base/math.sqrt(2))
hatched0.normal = hatched0(_base)
hatched0.large = hatched0(_base*math.sqrt(2))
hatched0.Large = hatched0(_base*math.sqrt(4))
hatched0.LArge = hatched0(_base*math.sqrt(8))
hatched0.LARge = hatched0(_base*math.sqrt(16))
hatched0.LARGe = hatched0(_base*math.sqrt(32))
hatched0.LARGE = hatched0(_base*math.sqrt(64))

hatched45 = hatched(_base, 45)
hatched45.SMALL = hatched45(_base/math.sqrt(64))
hatched45.SMALl = hatched45(_base/math.sqrt(32))
hatched45.SMAll = hatched45(_base/math.sqrt(16))
hatched45.SMall = hatched45(_base/math.sqrt(8))
hatched45.Small = hatched45(_base/math.sqrt(4))
hatched45.small = hatched45(_base/math.sqrt(2))
hatched45.normal = hatched45(_base)
hatched45.large = hatched45(_base*math.sqrt(2))
hatched45.Large = hatched45(_base*math.sqrt(4))
hatched45.LArge = hatched45(_base*math.sqrt(8))
hatched45.LARge = hatched45(_base*math.sqrt(16))
hatched45.LARGe = hatched45(_base*math.sqrt(32))
hatched45.LARGE = hatched45(_base*math.sqrt(64))

hatched90 = hatched(_base, 90)
hatched90.SMALL = hatched90(_base/math.sqrt(64))
hatched90.SMALl = hatched90(_base/math.sqrt(32))
hatched90.SMAll = hatched90(_base/math.sqrt(16))
hatched90.SMall = hatched90(_base/math.sqrt(8))
hatched90.Small = hatched90(_base/math.sqrt(4))
hatched90.small = hatched90(_base/math.sqrt(2))
hatched90.normal = hatched90(_base)
hatched90.large = hatched90(_base*math.sqrt(2))
hatched90.Large = hatched90(_base*math.sqrt(4))
hatched90.LArge = hatched90(_base*math.sqrt(8))
hatched90.LARge = hatched90(_base*math.sqrt(16))
hatched90.LARGe = hatched90(_base*math.sqrt(32))
hatched90.LARGE = hatched90(_base*math.sqrt(64))

hatched135 = hatched(_base, 135)
hatched135.SMALL = hatched135(_base/math.sqrt(64))
hatched135.SMALl = hatched135(_base/math.sqrt(32))
hatched135.SMAll = hatched135(_base/math.sqrt(16))
hatched135.SMall = hatched135(_base/math.sqrt(8))
hatched135.Small = hatched135(_base/math.sqrt(4))
hatched135.small = hatched135(_base/math.sqrt(2))
hatched135.normal = hatched135(_base)
hatched135.large = hatched135(_base*math.sqrt(2))
hatched135.Large = hatched135(_base*math.sqrt(4))
hatched135.LArge = hatched135(_base*math.sqrt(8))
hatched135.LARge = hatched135(_base*math.sqrt(16))
hatched135.LARGe = hatched135(_base*math.sqrt(32))
hatched135.LARGE = hatched135(_base*math.sqrt(64))


class crosshatched(pattern):
    def __init__(self, dist, angle, strokestyles=[]):
        pattern.__init__(self, painttype=1, tilingtype=1, xstep=dist, ystep=dist, bbox=None, trafo=trafo.rotate(angle))
        self.strokestyles = attr.mergeattrs([style.linewidth.THIN] + strokestyles)
        attr.checkattrs(self.strokestyles, [style.strokestyle])
        self.dist = dist
        self.angle = angle
        self.stroke(path.line_pt(0, 0, 0, unit.topt(dist)), self.strokestyles)
        self.stroke(path.line_pt(0, 0, unit.topt(dist), 0), self.strokestyles)

    def __call__(self, dist=None, angle=None, strokestyles=None):
        if dist is None:
            dist = self.dist
        if angle is None:
            angle = self.angle
        if strokestyles is None:
            strokestyles = self.strokestyles
        return crosshatched(dist, angle, strokestyles)

crosshatched0 = crosshatched(_base, 0)
crosshatched0.SMALL = crosshatched0(_base/math.sqrt(64))
crosshatched0.SMALl = crosshatched0(_base/math.sqrt(32))
crosshatched0.SMAll = crosshatched0(_base/math.sqrt(16))
crosshatched0.SMall = crosshatched0(_base/math.sqrt(8))
crosshatched0.Small = crosshatched0(_base/math.sqrt(4))
crosshatched0.small = crosshatched0(_base/math.sqrt(2))
crosshatched0.normal = crosshatched0
crosshatched0.large = crosshatched0(_base*math.sqrt(2))
crosshatched0.Large = crosshatched0(_base*math.sqrt(4))
crosshatched0.LArge = crosshatched0(_base*math.sqrt(8))
crosshatched0.LARge = crosshatched0(_base*math.sqrt(16))
crosshatched0.LARGe = crosshatched0(_base*math.sqrt(32))
crosshatched0.LARGE = crosshatched0(_base*math.sqrt(64))

crosshatched45 = crosshatched(_base, 45)
crosshatched45.SMALL = crosshatched45(_base/math.sqrt(64))
crosshatched45.SMALl = crosshatched45(_base/math.sqrt(32))
crosshatched45.SMAll = crosshatched45(_base/math.sqrt(16))
crosshatched45.SMall = crosshatched45(_base/math.sqrt(8))
crosshatched45.Small = crosshatched45(_base/math.sqrt(4))
crosshatched45.small = crosshatched45(_base/math.sqrt(2))
crosshatched45.normal = crosshatched45
crosshatched45.large = crosshatched45(_base*math.sqrt(2))
crosshatched45.Large = crosshatched45(_base*math.sqrt(4))
crosshatched45.LArge = crosshatched45(_base*math.sqrt(8))
crosshatched45.LARge = crosshatched45(_base*math.sqrt(16))
crosshatched45.LARGe = crosshatched45(_base*math.sqrt(32))
crosshatched45.LARGE = crosshatched45(_base*math.sqrt(64))


class PDFpattern(pdfwriter.PDFobject):

    def __init__(self, name, patterntype, painttype, tilingtype, bbox, xstep, ystep, trafo,
                 patternproc, writer, registry, patternregistry):
        self.patternregistry = patternregistry
        pdfwriter.PDFobject.__init__(self, "pattern", name)
        registry.addresource("Pattern", name, self)

        self.name = name
        self.patterntype = patterntype
        self.painttype = painttype
        self.tilingtype = tilingtype
        self.bbox = bbox
        self.xstep = xstep
        self.ystep = ystep
        self.trafo = trafo
        self.patternproc = patternproc

    def write(self, file, writer, registry):
        file.write("<<\n"
                   "/Type /Pattern\n"
                   "/PatternType %d\n" % self.patterntype)
        file.write("/PaintType %d\n" % self.painttype)
        file.write("/TilingType %d\n" % self.tilingtype)
        file.write("/BBox [%d %d %d %d]\n" % self.bbox.lowrestuple_pt())
        file.write("/XStep %f\n" % self.xstep)
        file.write("/YStep %f\n" % self.ystep)
        file.write("/Matrix %s\n" % str(self.trafo))
        file.write("/Resources ")
        self.patternregistry.writeresources(file)
        if writer.compress:
            import zlib
            content = zlib.compress(self.patternproc)
        else:
            content = self.patternproc

        file.write("/Length %i\n" % len(content))
        if writer.compress:
            file.write("/Filter /FlateDecode\n")
        file.write(">>\n"
                   "stream\n")
        file.write_bytes(content)
        file.write("endstream\n")


class SVGpattern(svgwriter.SVGresource):

    def __init__(self, pattern):
        self.svgid = "pattern%d" % id(pattern)
        super().__init__("pattern", self.svgid)
        self.pattern = pattern

    def output(self, xml, writer, registry):
        if self.pattern.patternbbox:
            patternbbox = self.pattern.patternbbox
        else:
            patternfile = io.BytesIO()
            patternxml = svgwriter.SVGGenerator(patternfile)
            patternbbox = bboxmodule.empty()
            patternxml.startSVGDocument()
            self.pattern.processSVG(patternxml, writer, svgwriter.context(), svgwriter.SVGregistry(), patternbbox)

        attrs = {"id": self.svgid, "patternUnits": "userSpaceOnUse"}
        llx, lly, urx, ury = patternbbox.highrestuple_pt()
        attrs["viewBox"] = "%g %g %g %g" % (llx, -ury, urx-llx, ury-lly)
        attrs["width"] = "%g" % (urx-llx)
        attrs["height"] = "%g" % (ury-lly)
        if self.pattern.patterntrafo:
            self.pattern.patterntrafo.processSVGattrs(attrs, self, svgwriter.context(), registry)
            attrs["patternTransform"] = attrs["transform"]
            del attrs["transform"]

        xml.startSVGElement("pattern", attrs)
        self.pattern.processSVG(xml, writer, svgwriter.context(), registry, bboxmodule.empty())
        xml.endSVGElement("pattern")




# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2002-2004, 2006, 2013 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2003-2006 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import binascii, colorsys, logging, math, struct
from . import attr, style, pdfwriter

logger = logging.getLogger("pyx")

# device-dependent (nonlinear) functions for color conversion
# UCRx : [0,1] -> [-1, 1] UnderColorRemoval (removes black from c, y, m)
# BG   : [0,1] -> [0, 1]  BlackGeneration (generate the black from the nominal k-value)
# as long as we have no further knowledge we define them linearly with constants 1
def _UCRc(x): return x
def _UCRm(x): return x
def _UCRy(x): return x
def _BG(x): return x

def set(UCRc=None, UCRm=None, UCRy=None, BG=None):
    global _UCRc
    global _UCRm
    global _UCRy
    global _BG

    if UCRc is not None:
        _UCRc = UCRc
    if UCRm is not None:
        _UCRm = UCRm
    if UCRy is not None:
        _UCRy = UCRy
    if BG is not None:
        _BG = BG


class color(attr.exclusiveattr, style.strokestyle, style.fillstyle):
    """base class for all colors"""
    def __init__(self):
        super().__init__(color)

    def processSVGattrs(self, attrs, writer, context, registry):
        if context.strokeattr:
            context.strokecolor = self.rgb().tohexstring()
        if context.fillattr:
            context.fillcolor = self.rgb().tohexstring()


clear = attr.clearclass(color)


class gray(color):

    """grey tones"""

    def __init__(self, g=0.0):
        super().__init__()
        if g<0 or g>1:
            raise ValueError("Value g out of range [0,1]")
        self.g = g

    def processPS(self, file, writer, context, registry):
        file.write("%f setgray\n" % self.g)

    def processPDF(self, file, writer, context, registry):
        if context.strokeattr:
            file.write("%f G\n" % self.g)
        if context.fillattr:
            file.write("%f g\n" % self.g)

    def cmyk(self):
        return cmyk(0, 0, 0, 1 - self.g)

    def gray(self):
        return gray(self.g)
    grey = gray

    def hsb(self):
        return hsb(0, 0, self.g)

    def rgb(self):
        return rgb(self.g, self.g, self.g)

    def colorspacestring(self):
        return "/DeviceGray"

    def to8bitbytes(self):
        return bytes((int(self.g*255),))

gray.black = gray(0.0)
gray.white = gray(1.0)
grey = gray


class rgb(color):

    """rgb colors"""

    def __init__(self, r=0.0, g=0.0, b=0.0):
        super().__init__()
        if r<0 or r>1:
            raise ValueError("Value r out of range [0,1]")
        if g<0 or g>1:
            raise ValueError("Value g out of range [0,1]")
        if b<0 or b>1:
            raise ValueError("Value b out of range [0,1]")
        self.r = r
        self.g = g
        self.b = b

    def processPS(self, file, writer, context, registry):
        file.write("%f %f %f setrgbcolor\n" % (self.r, self.g, self.b))

    def processPDF(self, file, writer, context, registry):
        if context.strokeattr:
            file.write("%f %f %f RG\n" % (self.r, self.g, self.b))
        if context.fillattr:
            file.write("%f %f %f rg\n" % (self.r, self.g, self.b))

    def cmyk(self):
        # conversion to cmy
        c, m, y = 1 - self.r, 1 - self.g, 1 - self.b
        # conversion from cmy to cmyk with device-dependent functions
        k = min([c, m, y])
        return cmyk(min(1, max(0, c - _UCRc(k))),
                    min(1, max(0, m - _UCRm(k))),
                    min(1, max(0, y - _UCRy(k))),
                    _BG(k))

    def gray(self):
        return gray(0.3*self.r + 0.59*self.g + 0.11*self.b)
    grey = gray

    def hsb(self):
        h, s, b = colorsys.rgb_to_hsv(self.r, self.g, self.b)
        return hsb(h, s, b)

    def rgb(self):
        return rgb(self.r, self.g, self.b)

    def colorspacestring(self):
        return "/DeviceRGB"

    def to8bitbytes(self):
        return struct.pack("BBB", int(self.r*255), int(self.g*255), int(self.b*255))

    def tohexstring(self, cssstrip=1, addhash=1):
        hexstring = binascii.b2a_hex(self.to8bitbytes()).decode('ascii')
        if cssstrip and hexstring[0] == hexstring[1] and hexstring[2] == hexstring[3] and hexstring[4] == hexstring[5]:
            hexstring = "".join([hexstring[0], hexstring[2], hexstring[4]])
        if addhash:
            hexstring = "#" + hexstring
        return hexstring


def rgbfromhexstring(hexstring):
    hexstring = hexstring.strip().lstrip("#")
    if len(hexstring) == 3:
        hexstring = "".join([hexstring[0], hexstring[0], hexstring[1], hexstring[1], hexstring[2], hexstring[2]])
    elif len(hexstring) != 6:
        raise ValueError("3 or 6 digit hex number expected (with optional leading hash character)")
    return rgb(*[value/255.0 for value in struct.unpack("BBB", binascii.a2b_hex(hexstring))])

rgb.red   = rgb(1, 0, 0)
rgb.green = rgb(0, 1, 0)
rgb.blue  = rgb(0, 0, 1)
rgb.white = rgb(1, 1, 1)
rgb.black = rgb(0, 0, 0)


class hsb(color):

    """hsb colors"""

    def __init__(self, h=0.0, s=0.0, b=0.0):
        super().__init__()
        if h<0 or h>1:
            raise ValueError("Value h out of range [0,1]")
        if s<0 or s>1:
            raise ValueError("Value s out of range [0,1]")
        if b<0 or b>1:
            raise ValueError("Value b out of range [0,1]")
        self.h = h
        self.s = s
        self.b = b

    def processPS(self, file, writer, context, registry):
        file.write("%f %f %f sethsbcolor\n" % (self.h, self.s, self.b))

    def processPDF(self, file, writer, context, registry):
        self.rgb().processPDF(file, writer, context, registry)

    def cmyk(self):
        return self.rgb().cmyk()

    def gray(self):
        return self.rgb().gray()
    grey = gray

    def hsb(self):
        return hsb(self.h, self.s, self.b)

    def rgb(self):
        r, g, b = colorsys.hsv_to_rgb(self.h, self.s, self.b)
        return rgb(r, g, b)

    def colorspacestring(self):
        raise RuntimeError("colorspace string not available for hsb colors")


class cmyk(color):

    """cmyk colors"""

    def __init__(self, c=0.0, m=0.0, y=0.0, k=0.0):
        super().__init__()
        if c<0 or c>1:
            raise ValueError("Value c out of range [0,1]")
        if m<0 or m>1:
            raise ValueError("Value m out of range [0,1]")
        if y<0 or y>1:
            raise ValueError("Value y out of range [0,1]")
        if k<0 or k>1:
            raise ValueError("Value k out of range [0,1]")
        self.c = c
        self.m = m
        self.y = y
        self.k = k

    def processPS(self, file, writer, context, registry):
        file.write("%f %f %f %f setcmykcolor\n" % (self.c, self.m, self.y, self.k))

    def processPDF(self, file, writer, context, registry):
        if context.strokeattr:
            file.write("%f %f %f %f K\n" % (self.c, self.m, self.y, self.k))
        if context.fillattr:
            file.write("%f %f %f %f k\n" % (self.c, self.m, self.y, self.k))

    def cmyk(self):
        return cmyk(self.c, self.m, self.y, self.k)

    def gray(self):
        return gray(1 - min([1, 0.3*self.c + 0.59*self.m + 0.11*self.y + self.k]))
    grey = gray

    def hsb(self):
        return self.rgb().hsb()

    def rgb(self):
        # conversion to cmy:
        c = min(1, self.c + self.k)
        m = min(1, self.m + self.k)
        y = min(1, self.y + self.k)
        # conversion from cmy to rgb:
        return rgb(1 - c, 1 - m, 1 - y)

    def colorspacestring(self):
        return "/DeviceCMYK"

    def to8bitbytes(self):
        return struct.pack("BBBB", int(self.c*255), int(self.m*255), int(self.y*255), int(self.k*255))

cmyk.GreenYellow    = cmyk(0.15, 0, 0.69, 0)
cmyk.Yellow         = cmyk(0, 0, 1, 0)
cmyk.Goldenrod      = cmyk(0, 0.10, 0.84, 0)
cmyk.Dandelion      = cmyk(0, 0.29, 0.84, 0)
cmyk.Apricot        = cmyk(0, 0.32, 0.52, 0)
cmyk.Peach          = cmyk(0, 0.50, 0.70, 0)
cmyk.Melon          = cmyk(0, 0.46, 0.50, 0)
cmyk.YellowOrange   = cmyk(0, 0.42, 1, 0)
cmyk.Orange         = cmyk(0, 0.61, 0.87, 0)
cmyk.BurntOrange    = cmyk(0, 0.51, 1, 0)
cmyk.Bittersweet    = cmyk(0, 0.75, 1, 0.24)
cmyk.RedOrange      = cmyk(0, 0.77, 0.87, 0)
cmyk.Mahogany       = cmyk(0, 0.85, 0.87, 0.35)
cmyk.Maroon         = cmyk(0, 0.87, 0.68, 0.32)
cmyk.BrickRed       = cmyk(0, 0.89, 0.94, 0.28)
cmyk.Red            = cmyk(0, 1, 1, 0)
cmyk.OrangeRed      = cmyk(0, 1, 0.50, 0)
cmyk.RubineRed      = cmyk(0, 1, 0.13, 0)
cmyk.WildStrawberry = cmyk(0, 0.96, 0.39, 0)
cmyk.Salmon         = cmyk(0, 0.53, 0.38, 0)
cmyk.CarnationPink  = cmyk(0, 0.63, 0, 0)
cmyk.Magenta        = cmyk(0, 1, 0, 0)
cmyk.VioletRed      = cmyk(0, 0.81, 0, 0)
cmyk.Rhodamine      = cmyk(0, 0.82, 0, 0)
cmyk.Mulberry       = cmyk(0.34, 0.90, 0, 0.02)
cmyk.RedViolet      = cmyk(0.07, 0.90, 0, 0.34)
cmyk.Fuchsia        = cmyk(0.47, 0.91, 0, 0.08)
cmyk.Lavender       = cmyk(0, 0.48, 0, 0)
cmyk.Thistle        = cmyk(0.12, 0.59, 0, 0)
cmyk.Orchid         = cmyk(0.32, 0.64, 0, 0)
cmyk.DarkOrchid     = cmyk(0.40, 0.80, 0.20, 0)
cmyk.Purple         = cmyk(0.45, 0.86, 0, 0)
cmyk.Plum           = cmyk(0.50, 1, 0, 0)
cmyk.Violet         = cmyk(0.79, 0.88, 0, 0)
cmyk.RoyalPurple    = cmyk(0.75, 0.90, 0, 0)
cmyk.BlueViolet     = cmyk(0.86, 0.91, 0, 0.04)
cmyk.Periwinkle     = cmyk(0.57, 0.55, 0, 0)
cmyk.CadetBlue      = cmyk(0.62, 0.57, 0.23, 0)
cmyk.CornflowerBlue = cmyk(0.65, 0.13, 0, 0)
cmyk.MidnightBlue   = cmyk(0.98, 0.13, 0, 0.43)
cmyk.NavyBlue       = cmyk(0.94, 0.54, 0, 0)
cmyk.RoyalBlue      = cmyk(1, 0.50, 0, 0)
cmyk.Blue           = cmyk(1, 1, 0, 0)
cmyk.Cerulean       = cmyk(0.94, 0.11, 0, 0)
cmyk.Cyan           = cmyk(1, 0, 0, 0)
cmyk.ProcessBlue    = cmyk(0.96, 0, 0, 0)
cmyk.SkyBlue        = cmyk(0.62, 0, 0.12, 0)
cmyk.Turquoise      = cmyk(0.85, 0, 0.20, 0)
cmyk.TealBlue       = cmyk(0.86, 0, 0.34, 0.02)
cmyk.Aquamarine     = cmyk(0.82, 0, 0.30, 0)
cmyk.BlueGreen      = cmyk(0.85, 0, 0.33, 0)
cmyk.Emerald        = cmyk(1, 0, 0.50, 0)
cmyk.JungleGreen    = cmyk(0.99, 0, 0.52, 0)
cmyk.SeaGreen       = cmyk(0.69, 0, 0.50, 0)
cmyk.Green          = cmyk(1, 0, 1, 0)
cmyk.ForestGreen    = cmyk(0.91, 0, 0.88, 0.12)
cmyk.PineGreen      = cmyk(0.92, 0, 0.59, 0.25)
cmyk.LimeGreen      = cmyk(0.50, 0, 1, 0)
cmyk.YellowGreen    = cmyk(0.44, 0, 0.74, 0)
cmyk.SpringGreen    = cmyk(0.26, 0, 0.76, 0)
cmyk.OliveGreen     = cmyk(0.64, 0, 0.95, 0.40)
cmyk.RawSienna      = cmyk(0, 0.72, 1, 0.45)
cmyk.Sepia          = cmyk(0, 0.83, 1, 0.70)
cmyk.Brown          = cmyk(0, 0.81, 1, 0.60)
cmyk.Tan            = cmyk(0.14, 0.42, 0.56, 0)
cmyk.Gray           = cmyk(0, 0, 0, 0.50)
cmyk.Grey           = cmyk.Gray
cmyk.Black          = cmyk(0, 0, 0, 1)
cmyk.White          = cmyk(0, 0, 0, 0)
cmyk.white          = cmyk.White
cmyk.black          = cmyk.Black


class palette(attr.changelist):
    """color palettes

    A color palette is a discrete, ordered list of colors"""

palette.clear = attr.clearclass(palette)

#
# gradients
#

class gradient(attr.changeattr):

    """base class for color gradients

    A gradient is a continuous collection of colors with a single parameter ranging from 0 to 1
    to address them"""

    def getcolor(self, param):
        """return color corresponding to param"""
        pass

    def select(self, index, n_indices):
        """return a color corresponding to an index out of n_indices"""
        if n_indices == 1:
            param = 0
        else:
            param = index / (n_indices - 1.0)
        return self.getcolor(param)

gradient.clear = attr.clearclass(gradient)

#
# gradient with arbitrary non-linear dependency
#

class functiongradient_gray(gradient):

    """arbitrary non-linear gradients of gray colors

    f_gray: a function mapping [0,1] to the gray value
    """

    def __init__(self, f_gray):
        super().__init__()
        self.f_gray = f_gray

    def getcolor(self, param):
        return gray(self.f_gray(param))


class functiongradient_cmyk(gradient):

    """arbitrary non-linear gradients of cmyk colors

    f_c: a function mapping [0,1] to the c component
    f_m: a function mapping [0,1] to the m component
    f_y: a function mapping [0,1] to the y component
    f_k: a function mapping [0,1] to the k component
    """

    def __init__(self, f_c, f_m, f_y, f_k):
        super().__init__()
        self.f_c = f_c
        self.f_m = f_m
        self.f_y = f_y
        self.f_k = f_k

    def getcolor(self, param):
        return cmyk(self.f_c(param), self.f_m(param), self.f_y(param), self.f_k(param))


class functiongradient_hsb(gradient):

    """arbitrary non-linear gradients of hsb colors

    f_h: a function mapping [0,1] to the h component
    f_s: a function mapping [0,1] to the s component
    f_b: a function mapping [0,1] to the b component
    """

    def __init__(self, f_h, f_s, f_b):
        super().__init__()
        self.f_h = f_h
        self.f_s = f_s
        self.f_b = f_b

    def getcolor(self, param):
        return hsb(self.f_h(param), self.f_s(param), self.f_b(param))


class functiongradient_rgb(gradient):

    """arbitrary non-linear gradients of rgb colors

    f_r: a function mapping [0,1] to the r component
    f_g: a function mapping [0,1] to the b component
    f_b: a function mapping [0,1] to the b component
    """

    def __init__(self, f_r, f_g, f_b):
        super().__init__()
        self.f_r = f_r
        self.f_g = f_g
        self.f_b = f_b

    def getcolor(self, param):
        return rgb(self.f_r(param), self.f_g(param), self.f_b(param))

#
# factory functions for gradients interpolating linearly between two colors
#

def lineargradient_cmyk(mincolor, maxcolor):
     return functiongradient_cmyk(lambda x:maxcolor.c * x + mincolor.c * (1-x),
                                  lambda x:maxcolor.m * x + mincolor.m * (1-x),
                                  lambda x:maxcolor.y * x + mincolor.y * (1-x),
                                  lambda x:maxcolor.k * x + mincolor.k * (1-x))

def lineargradient_gray(mincolor, maxcolor):
     return functiongradient_gray(lambda x:maxcolor.g * x + mincolor.g * (1-x))

def lineargradient_hsb(mincolor, maxcolor):
     return functiongradient_hsb(lambda x:maxcolor.h * x + mincolor.h * (1-x),
                                 lambda x:maxcolor.s * x + mincolor.s * (1-x),
                                 lambda x:maxcolor.b * x + mincolor.b * (1-x))

def lineargradient_rgb(mincolor, maxcolor):
     return functiongradient_rgb(lambda x:maxcolor.r * x + mincolor.r * (1-x),
                                 lambda x:maxcolor.g * x + mincolor.g * (1-x),
                                 lambda x:maxcolor.b * x + mincolor.b * (1-x))


#
# gradients converted into other color spaces
#

class rgbgradient(gradient):

    "a gradient, which takes another gradient and returns rgb colors"

    def __init__(self, gradient):
        super().__init__()
        self.gradient = gradient

    def getcolor(self, param):
        return self.gradient.getcolor(param).rgb()


class cmykgradient(gradient):

    "a gradient, which takes another gradient and returns cmyk colors"

    def __init__(self, gradient):
        super().__init__()
        self.gradient = gradient

    def getcolor(self, param):
        return self.gradient.getcolor(param).cmyk()


gradient.Gray           = lineargradient_gray(gray.white, gray.black)
gradient.Grey           = gradient.Gray
gradient.ReverseGray    = lineargradient_gray(gray.black, gray.white)
gradient.ReverseGrey    = gradient.ReverseGray
gradient.BlackYellow    = functiongradient_rgb( # compare this with reversegray above
    f_r=lambda x: 2*x*(1-x)**5 + 3.5*x**2*(1-x)**3 + 2.1*x*x*(1-x)**2 + 3.0*x**3*(1-x)**2 + x**0.5*(1-(1-x)**2),
    f_g=lambda x: 1.5*x**2*(1-x)**3 - 0.8*x**3*(1-x)**2 + 2.0*x**4*(1-x) + x**4,
    f_b=lambda x: 5*x*(1-x)**5 - 0.5*x**2*(1-x)**3 + 0.3*x*x*(1-x)**2 + 5*x**3*(1-x)**2 + 0.5*x**6)
gradient.YellowBlack    = functiongradient_rgb(
    f_r=lambda x: 2*(1-x)*x**5 + 3.5*(1-x)**2*x**3 + 2.1*(1-x)*(1-x)*x**2 + 3.0*(1-x)**3*x**2 + (1-x)**0.5*(1-x**2),
    f_g=lambda x: 1.5*(1-x)**2*x**3 - 0.8*(1-x)**3*x**2 + 2.0*(1-x)**4*x + (1-x)**4,
    f_b=lambda x: 5*(1-x)*x**5 - 0.5*(1-x)**2*x**3 + 0.3*(1-x)*(1-x)*x**2 + 5*(1-x)**3*x**2 + 0.5*(1-x)**6)
gradient.RedGreen       = lineargradient_rgb(rgb.red, rgb.green)
gradient.RedBlue        = lineargradient_rgb(rgb.red, rgb.blue)
gradient.GreenRed       = lineargradient_rgb(rgb.green, rgb.red)
gradient.GreenBlue      = lineargradient_rgb(rgb.green, rgb.blue)
gradient.BlueRed        = lineargradient_rgb(rgb.blue, rgb.red)
gradient.BlueGreen      = lineargradient_rgb(rgb.blue, rgb.green)
gradient.RedBlack       = lineargradient_rgb(rgb.red, rgb.black)
gradient.BlackRed       = lineargradient_rgb(rgb.black, rgb.red)
gradient.RedWhite       = lineargradient_rgb(rgb.red, rgb.white)
gradient.WhiteRed       = lineargradient_rgb(rgb.white, rgb.red)
gradient.GreenBlack     = lineargradient_rgb(rgb.green, rgb.black)
gradient.BlackGreen     = lineargradient_rgb(rgb.black, rgb.green)
gradient.GreenWhite     = lineargradient_rgb(rgb.green, rgb.white)
gradient.WhiteGreen     = lineargradient_rgb(rgb.white, rgb.green)
gradient.BlueBlack      = lineargradient_rgb(rgb.blue, rgb.black)
gradient.BlackBlue      = lineargradient_rgb(rgb.black, rgb.blue)
gradient.BlueWhite      = lineargradient_rgb(rgb.blue, rgb.white)
gradient.WhiteBlue      = lineargradient_rgb(rgb.white, rgb.blue)
gradient.Rainbow        = lineargradient_hsb(hsb(0, 1, 1), hsb(2.0/3.0, 1, 1))
gradient.ReverseRainbow = lineargradient_hsb(hsb(2.0/3.0, 1, 1), hsb(0, 1, 1))
gradient.Hue            = lineargradient_hsb(hsb(0, 1, 1), hsb(1, 1, 1))
gradient.ReverseHue     = lineargradient_hsb(hsb(1, 1, 1), hsb(0, 1, 1))
rgbgradient.Rainbow        = rgbgradient(gradient.Rainbow)
rgbgradient.ReverseRainbow = rgbgradient(gradient.ReverseRainbow)
rgbgradient.Hue            = rgbgradient(gradient.Hue)
rgbgradient.ReverseHue     = rgbgradient(gradient.ReverseHue)
cmykgradient.Rainbow        = cmykgradient(gradient.Rainbow)
cmykgradient.ReverseRainbow = cmykgradient(gradient.ReverseRainbow)
cmykgradient.Hue            = cmykgradient(gradient.Hue)
cmykgradient.ReverseHue     = cmykgradient(gradient.ReverseHue)
def jet_r(x):
    if x < 0.38: return 0
    elif x < 0.62: return (x-0.38)/(0.62-0.38)
    elif x < 0.87: return 1
    else: return 0.5 + 0.5*(1-x)/(1-0.87)
def jet_g(x):
    if x < 0.13: return 0
    elif x < 0.38: return (x-0.13)/(0.38-0.13)
    elif x < 0.62: return 1
    elif x < 0.87: return (0.87-x)/(0.87-0.62)
    else: return 0
def jet_b(x):
    if x < 0.13: return 0.5 + 0.5*x/0.13
    elif x < 0.38: return 1
    elif x < 0.62: return 1-(x-0.38)/(0.62-0.38)
    else: return 0
gradient.Jet = functiongradient_rgb(f_r=jet_r, f_g=jet_g, f_b=jet_b)
gradient.ReverseJet = functiongradient_rgb(f_r=lambda x: jet_r(1-x), f_g=lambda x: jet_g(1-x), f_b=lambda x: jet_b(1-x))
cmykgradient.Jet = cmykgradient(gradient.Jet)
cmykgradient.ReverseJet = cmykgradient(gradient.ReverseJet)



class PDFextgstate(pdfwriter.PDFobject):

    def __init__(self, name, extgstate, registry):
        pdfwriter.PDFobject.__init__(self, "extgstate", name)
        registry.addresource("ExtGState", name, self)
        self.name = name
        self.extgstate = extgstate

    def write(self, file, writer, registry):
        file.write("%s\n" % self.extgstate)


class transparency(attr.exclusiveattr, style.strokestyle, style.fillstyle):

    def __init__(self, value):
        self.value = 1-value
        attr.exclusiveattr.__init__(self, transparency)

    def processPS(self, file, writer, context, registry):
        logger.warning("Transparency not available in PostScript, proprietary ghostscript extension code inserted.")
        file.write("%f .setshapealpha\n" % self.value)

    def processPDF(self, file, writer, context, registry):
        if context.strokeattr and context.fillattr:
            registry.add(PDFextgstate("Transparency-%f" % self.value,
                                      "<< /Type /ExtGState /CA %f /ca %f >>" % (self.value, self.value), registry))
            file.write("/Transparency-%f gs\n" % self.value)
        elif context.strokeattr:
            registry.add(PDFextgstate("Transparency-Stroke-%f" % self.value,
                                      "<< /Type /ExtGState /CA %f >>" % self.value, registry))
            file.write("/Transparency-Stroke-%f gs\n" % self.value)
        elif context.fillattr:
            registry.add(PDFextgstate("Transparency-Fill-%f" % self.value,
                                      "<< /Type /ExtGState /ca %f >>" % self.value, registry))
            file.write("/Transparency-Fill-%f gs\n" % self.value)

    def processSVGattrs(self, attrs, writer, context, registry):
        if context.strokeattr:
            context.strokeopacity = self.value
        if context.fillattr:
            context.fillopacity = self.value




from pyx import *
unit.set(xscale=3)

c = canvas.canvas()
t = c.text(0, 0, r"Here\PyXMarker{id} is a marker.")
center = t.marker("id")
c.stroke(path.circle(center[0], center[1], 1), [color.rgb.red])
c.stroke(path.line(center[0]-1, center[1], center[0]+1, center[1]), [color.rgb.red])
c.stroke(path.line(center[0], center[1]-1, center[0], center[1]+1), [color.rgb.red])
c.writeEPSfile("marker")
c.writePDFfile("marker")
c.writeSVGfile("marker")



from pyx import *

c = canvas.canvas()
p = path.line(0, 0, 2, 2)
p.append(path.curveto(2, 0, 3, 0, 4, 0))
c.stroke(p)
c.stroke(p, [deformer.parallel(0.2), color.rgb.blue])
c.stroke(p, [deformer.parallel(-0.2), color.rgb.red])
c.writeEPSfile("parallel")
c.writePDFfile("parallel")
c.writeSVGfile("parallel")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2005-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from . import font
__allfont__ = ["T1font", "T1builtinfont", "text_pt", "font"]
for importfromfont in __allfont__:
    locals()[importfromfont] = getattr(font, importfromfont)

__all__ = __allfont__



from pyx import *
from PIL import Image

im = Image.new("RGB", (3, 1))
im.putpixel((0, 0), (255, 0, 0))
im.putpixel((1, 0), (0, 255, 0))
im.putpixel((2, 0), (0, 0, 255))

c = canvas.canvas()
c.insert(bitmap.bitmap(0, 0, im, height=0.8))
c.writeEPSfile("pil")
c.writePDFfile("pil")
c.writeSVGfile("pil")



from pyx import *

g = graph.graphxy(width=8, x=graph.axis.bar())
g.plot(graph.data.file("minimal.dat", xname=0, y=2), [graph.style.changebar()])
g.writeEPSfile("changebar")
g.writePDFfile("changebar")
g.writeSVGfile("changebar")



from pyx import *

# colors and symbols to use (alternatingly)
colors = [color.rgb.red, color.rgb.green, color.rgb.blue, color.gray.black]
symbols = [graph.style._diamondsymbol, graph.style._trianglesymbol, graph.style._circlesymbol]

# create the graph styles to be used below
symbol = graph.style.symbol(symbol=attr.changelist(symbols),
                            symbolattrs=[deco.stroked.clear,
                                         attr.changelist([deco.filled([cc]) for cc in colors])])
line = graph.style.line(lineattrs=[attr.changelist(colors),
                                   attr.changelist([style.linestyle.solid])])

g = graph.graphxy(width=8, x=graph.axis.linear(min=0, max=1))
g.plot([graph.data.function("y(x) = x**%d" % i, points=8) for i in range(1, 7)],
       styles=[line, symbol])
g.writeEPSfile("usesymbol")
g.writePDFfile("usesymbol")
g.writeSVGfile("usesymbol")



from pyx import *

p = path.curve(0, 0, 1, 0, 1, 1, 2, 1)
p1, p2, p3 = p.split([0.5, p.end()-0.5])

c = canvas.canvas()
c.stroke(p, [style.linewidth.Thin, color.rgb.red])
c.stroke(p2, [style.linewidth.Thick, color.rgb.green])
c.writeEPSfile("split")
c.writePDFfile("split")
c.writeSVGfile("split")



from pyx import *

p2 = path.curve(0, 0, 3, 0, 1, 4, 4, 4)
p1 = p2.transformed(trafo.translate(-4, 0).scaled(0.75))
p3 = p2.transformed(trafo.scale(1.25).translated(4, 0))

myaxis = graph.axis.linear(min=0, max=10)

c = canvas.canvas()
c.insert(graph.axis.pathaxis(p1, myaxis))
c.insert(graph.axis.pathaxis(p2, myaxis))
c.insert(graph.axis.pathaxis(p3, myaxis))
c.writeEPSfile("rating")
c.writePDFfile("rating")
c.writeSVGfile("rating")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2005-2011 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2005-2011 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import io, copy, time, math, os
from . import bbox, config, style, version, unit, trafo, writer


class PSregistry:

    def __init__(self):
        # in order to keep a consistent order of the registered resources we
        # not only store them in a hash but also keep an ordered list (up to a
        # possible merging of resources, in which case the first instance is
        # kept)
        self.resourceshash = {}
        self.resourceslist = []

    def add(self, resource):
        rkey = (resource.type, resource.id)
        if rkey in self.resourceshash:
           self.resourceshash[rkey].merge(resource)
        else:
           self.resourceshash[rkey] = resource
           self.resourceslist.append(resource)

    def mergeregistry(self, registry):
        for resource in registry.resources:
            self.add(resource)

    def output(self, file, writer):
        """ write all PostScript code of the prolog resources """
        for resource in self.resourceslist:
            resource.output(file, writer, self)

#
# Abstract base class
#

class PSresource:

    """ a PostScript resource """

    def __init__(self, type, id):
        # Every PSresource has to have a type and a unique id.
        # Resources with the same type and id will be merged
        # when they are registered in the PSregistry
        self.type = type
        self.id = id

    def merge(self, other):
        """ merge self with other, which has to be a resource of the same type and with
        the same id"""
        pass

    def output(self, file, writer, registry):
        raise NotImplementedError("output not implemented for %s" % repr(self))

class PSdefinition(PSresource):

    """ PostScript function definition included in the prolog """

    def __init__(self, id, body):
        self.type = "definition"
        self.id = id
        self.body = body

    def output(self, file, writer, registry):
        file.write("%%%%BeginResource: %s\n" % self.id)
        file.write_bytes(self.body)
        file.write(" /%s exch def\n" % self.id)
        file.write("%%EndResource\n")

#
# Writers
#

class _PSwriter:

    def __init__(self, title=None, 
                 stripfonts=True, textaspath=False, meshasbitmap=False, meshasbitmapresolution=300,
                 strip_fonts=None, text_as_path=None, mesh_as_bitmap=None, mesh_as_bitmap_resolution=None):
        self._fontmap = None
        self.title = title
        if strip_fonts is not None:
            logger.warning("(E)PSwriter: strip_fonts deprecated, use stripfonts instead")
            stripfonts = strip_fonts
        self.stripfonts = stripfonts
        if text_as_path is not None:
            logger.warning("(E)PSwriter: text_as_path deprecated, use textaspath instead")
            textaspath = text_as_path
        self.textaspath = textaspath
        if mesh_as_bitmap is not None:
            logger.warning("(E)PSwriter: mesh_as_bitmap deprecated, use meshasbitmap instead")
            meshasbitmap = mash_as_bitmap
        self.meshasbitmap = meshasbitmap
        if mesh_as_bitmap_resolution is not None:
            logger.warning("PSwriter: mesh_as_bitmap_resolution deprecated, use meshasbitmapresolution instead")
            meshasbitmapresolution = mash_as_bitmap_resolution

        # dictionary mapping font names to dictionaries mapping encoding names to encodings
        # encodings themselves are mappings from glyphnames to codepoints
        self.encodings = {}

    def writeinfo(self, file):
        if os.environ.get('SOURCE_DATE_EPOCH'):
            creation_date = time.gmtime(int(os.environ.get('SOURCE_DATE_EPOCH')))
        else:
            creation_date = time.localtime()
        file.write("%%%%Creator: PyX %s\n" % version.version)
        if self.title is not None:
            file.write("%%%%Title: %s\n" % self.title)
        file.write("%%%%CreationDate: %s\n" %
                   time.asctime(creation_date))

    def getfontmap(self):
        if self._fontmap is None:
            # late import due to cyclic dependency
            from pyx.dvi import mapfile
            fontmapfiles = config.getlist("text", "psfontmaps", ["psfonts.map"])
            self._fontmap = mapfile.readfontmap(fontmapfiles)
        return self._fontmap


class EPSwriter(_PSwriter):

    def __init__(self, document, file, **kwargs):
        _PSwriter.__init__(self, **kwargs)
        file = writer.writer(file)

        if len(document.pages) != 1:
            raise ValueError("EPS file can be constructed out of a single page document only")
        page = document.pages[0]
        canvas = page.canvas

        pagefile = writer.writer(io.BytesIO())
        registry = PSregistry()
        acontext = context()
        pagebbox = bbox.empty()

        page.processPS(pagefile, self, acontext, registry, pagebbox)

        file.write("%!PS-Adobe-3.0 EPSF-3.0\n")
        if pagebbox:
            file.write("%%%%BoundingBox: %d %d %d %d\n" % pagebbox.lowrestuple_pt())
            file.write("%%%%HiResBoundingBox: %g %g %g %g\n" % pagebbox.highrestuple_pt())
        self.writeinfo(file)
        file.write("%%EndComments\n")

        file.write("%%BeginProlog\n")
        registry.output(file, self)
        file.write("%%EndProlog\n")

        file.write_bytes(pagefile.file.getvalue())

        file.write("showpage\n")
        file.write("%%Trailer\n")
        file.write("%%EOF\n")


class PSwriter(_PSwriter):

    def __init__(self, document, file, writebbox=False, **kwargs):
        _PSwriter.__init__(self, **kwargs)
        file = writer.writer(file)

        # We first have to process the content of the pages, writing them into the stream pagesfile
        # Doing so, we fill the registry and also calculate the page bounding boxes, which are
        # stored in page._bbox for every page
        pagesfile = writer.writer(io.BytesIO())
        registry = PSregistry()

        # calculated bounding boxes of the whole document
        documentbbox = bbox.empty()

        for nr, page in enumerate(document.pages):
            # process contents of page
            pagefile = writer.writer(io.BytesIO())
            acontext = context()
            pagebbox = bbox.empty()
            page.processPS(pagefile, self, acontext, registry, pagebbox)

            documentbbox += pagebbox

            pagesfile.write("%%%%Page: %s %d\n" % (page.pagename is None and str(nr+1) or page.pagename, nr+1))
            if page.paperformat:
                pagesfile.write("%%%%PageMedia: %s\n" % page.paperformat.name)
            pagesfile.write("%%%%PageOrientation: %s\n" % (page.rotated and "Landscape" or "Portrait"))
            if pagebbox and writebbox:
                pagesfile.write("%%%%PageBoundingBox: %d %d %d %d\n" % pagebbox.lowrestuple_pt())

            # page setup section
            pagesfile.write("%%BeginPageSetup\n")
            pagesfile.write("/pgsave save def\n")

            pagesfile.write("%%EndPageSetup\n")
            pagesfile.write_bytes(pagefile.file.getvalue())
            pagesfile.write("pgsave restore\n")
            pagesfile.write("showpage\n")
            pagesfile.write("%%PageTrailer\n")

        file.write("%!PS-Adobe-3.0\n")
        if documentbbox and writebbox:
            file.write("%%%%BoundingBox: %d %d %d %d\n" % documentbbox.lowrestuple_pt())
            file.write("%%%%HiResBoundingBox: %g %g %g %g\n" % documentbbox.highrestuple_pt())
        self.writeinfo(file)

        # required paper formats
        paperformats = {}
        for page in document.pages:
            if page.paperformat:
                paperformats[page.paperformat] = page.paperformat

        first = 1
        for paperformat in list(paperformats.values()):
            if first:
                file.write("%%DocumentMedia: ")
                first = 0
            else:
                file.write("%%+ ")
            file.write("%s %d %d 75 white ()\n" % (paperformat.name,
                                                   unit.topt(paperformat.width),
                                                   unit.topt(paperformat.height)))

        # file.write(%%DocumentNeededResources: ") # register not downloaded fonts here

        file.write("%%%%Pages: %d\n" % len(document.pages))
        file.write("%%PageOrder: Ascend\n")
        file.write("%%EndComments\n")

        # document defaults section
        #file.write("%%BeginDefaults\n")
        #file.write("%%EndDefaults\n")

        # document prolog section
        file.write("%%BeginProlog\n")
        registry.output(file, self)
        file.write("%%EndProlog\n")

        # document setup section
        #file.write("%%BeginSetup\n")
        #file.write("%%EndSetup\n")

        file.write_bytes(pagesfile.file.getvalue())

        file.write("%%Trailer\n")
        file.write("%%EOF\n")


class context:

    def __init__(self):
        self.linewidth_pt = None
        self.colorspace = None
        self.selectedfont = None
        self.fillrule = 0

    def __call__(self, **kwargs):
        newcontext = copy.copy(self)
        for key, value in list(kwargs.items()):
            setattr(newcontext, key, value)
        return newcontext



import math
from pyx import *

p1 = path.curve(0, 0, 3, 0, 1, 4, 4, 4)
p2 = p1.transformed(trafo.translate(4, 0))

myticks = [graph.axis.tick.tick(math.pi, label=r"\pi", labelattrs=[text.mathmode]),
           graph.axis.tick.tick(2*math.pi, label=r"2\pi", labelattrs=[text.mathmode])]

c = canvas.canvas()
c.insert(graph.axis.pathaxis(p1, graph.axis.linear(min=0, max=10)))
c.insert(graph.axis.pathaxis(p2, graph.axis.linear(min=0, max=10, manualticks=myticks)))
c.writeEPSfile("manualticks")
c.writePDFfile("manualticks")
c.writeSVGfile("manualticks")



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2017, 2021, 2022 Jörg Lehmann <joerg@pyx-project.org>
# Copyright (C) 2017, 2021, 2022 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

def kwsplit(kwargs, parts=None, allow_only_split=False):
    """
    .. testsetup::

       from pyx.utils import * 

    >>> kwsplit({"aaa": "bbb"})
    ({}, {'aaa': 'bbb'})

    >>> kwsplit({"aaa": "bbb"}, allow_only_split=True)
    Traceback (most recent call last):
    ...
    ValueError: Invalid kwargs: {'aaa': 'bbb'}

    >>> kwsplit({"aaa": "bbb"}, ["ccc"])
    ({}, {'aaa': 'bbb'})

    >>> kwsplit({"aaa": "bbb"}, ["aaa"])
    ({}, {'aaa': 'bbb'})

    >>> kwsplit({"aaa_bbb": "ccc"})
    ({'aaa': {'bbb': 'ccc'}}, {})

    >>> kwsplit({"aaa_bbb": "ccc"}, ["aaa"])
    ({'aaa': {'bbb': 'ccc'}}, {})

    >>> kwsplit({"aaa_bbb": "ccc"}, ["ddd"])
    ({}, {'aaa_bbb': 'ccc'})

    >>> kwsplit({"aaa_bbb": "ccc", "ddd_eee": "fff", "ggg": "hhh"})
    ({'aaa': {'bbb': 'ccc'}, 'ddd': {'eee': 'fff'}}, {'ggg': 'hhh'})

    >>> kwsplit({"aaa_bbb": "ccc", "ddd_eee": "fff", "ggg": "hhh"}, ["aaa"])
    ({'aaa': {'bbb': 'ccc'}}, {'ddd_eee': 'fff', 'ggg': 'hhh'})

    """
    split = {}
    rest = {}
    for key, value in kwargs.items():
        try:
            part, subkey = key.split("_", 1)
        except ValueError:
            rest[key] = value
        else:
            if parts is None or part in parts:
                if part not in split:
                    split[part] = {}
                split[part][subkey] = value
            else:
                rest[key] = value
    if allow_only_split and rest:
        raise ValueError("Invalid kwargs: %s" % rest)
    return split, rest


def merge_members_kwargs(obj, kwargs_updates, member_names):
    """return merge obj's member variables member_names and obj.kwargs with dictionary kwargs"""

    new_kwargs = {}
    for name in member_names:
        new_kwargs[name] = getattr(obj, name)

    for kwargs in kwargs_updates:
        new_kwargs.update(kwargs)
    return new_kwargs



# -*- encoding: utf-8 -*-
#
#
# Copyright (C) 2005-2006 André Wobst <wobsta@pyx-project.org>
#
# This file is part of PyX (https://pyx-project.org/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

import array

c1 = 52845
c1_16, c1_8 = divmod(c1, 0x100) # to avoid overflow (or conversion to the slow long integers)
c2 = 22719

def decoder(code, r, n):
    plain = array.array("B")
    for x in array.array("B", code):
        plain.append(x ^ (r >> 8))
        # r = ((x + r) * c1 + c2) & 0xffff # this might overflow
        r = ((((x + r) * c1_16) & 0xff) * 0x100 + (x + r) * c1_8 + c2) & 0xffff
    return plain.tobytes()[n:]

def encoder(data, r, random):
    code = array.array("B")
    for x in array.array("B", random+data):
        code.append(x ^ (r>>8))
        # r = ((code[-1] + r) * c1 + c2) & 0xffff # this might overflow
        r = ((((code[-1] + r) * c1_16) & 0xff) * 0x100 + (code[-1] + r) * c1_8 + c2) & 0xffff
    return code.tobytes()



import math
from pyx import *

p = path.curve(0, 0, 3, 0, 1, 4, 4, 4)

myparter = graph.axis.parter.linear(["1/3", "1/6"])

c = canvas.canvas()
c.insert(graph.axis.pathaxis(p, graph.axis.linear(min=0, max=1, parter=myparter)))
c.insert(graph.axis.pathaxis(p.transformed(trafo.translate(4, 0)),
                             graph.axis.linear(min=0, max=1, parter=myparter,
                             texter=graph.axis.texter.rational())))
c.writeEPSfile("parter")
c.writePDFfile("parter")
c.writeSVGfile("parter")
